'''
Created on Jul 3, 2014

@author: roj-idl71
'''
# SUBCHANNELS EN VEZ DE CHANNELS
# BENCHMARKS -> PROBLEMAS CON ARCHIVOS GRANDES -> INCONSTANTE EN EL TIEMPO
# ACTUALIZACION DE VERSION
# HEADERS
# MODULO DE ESCRITURA
# METADATA

import os
import time
import datetime
import numpy
import timeit
from fractions import Fraction
from time import time
from time import sleep

import schainpy.admin
from schainpy.model.data.jroheaderIO import RadarControllerHeader, SystemHeader
from schainpy.model.data.jrodata import Voltage
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator

import pickle
try:
    import digital_rf
except:
    pass


class DigitalRFReader(ProcessingUnit):
    '''
    classdocs
    '''

    def __init__(self):
        '''
        Constructor
        '''

        ProcessingUnit.__init__(self)

        self.dataOut                  = Voltage()
        self.__printInfo              = True
        self.__flagDiscontinuousBlock = False
        self.__bufferIndex = 9999999
        self.__codeType    = 0
        self.__ippKm       = None
        self.__nCode       = None
        self.__nBaud       = None
        self.__code        = None
        self.dtype         = None
        self.oldAverage    = None
        self.path          = None

    def close(self):
        print('Average of writing to digital rf format is ', self.oldAverage * 1000)
        return

    def __getCurrentSecond(self):

        return self.__thisUnixSample / self.__sample_rate

    thisSecond = property(__getCurrentSecond, "I'm the 'thisSecond' property.")

    def __setFileHeader(self):
        '''
        In this method will be initialized every parameter of dataOut object (header, no data)
        '''
        ippSeconds = 1.0 * self.__nSamples / self.__sample_rate

        nProfiles = 1.0 / ippSeconds  # Number of profiles in one second

        try:
            self.dataOut.radarControllerHeaderObj = RadarControllerHeader(
                self.__radarControllerHeader)
        except:
            self.dataOut.radarControllerHeaderObj = RadarControllerHeader(
                txA=0,
                txB=0,
                nWindows=1,
                nHeights=self.__nSamples,
                firstHeight=self.__firstHeigth,
                deltaHeight=self.__deltaHeigth,
                codeType=self.__codeType,
                nCode=self.__nCode, nBaud=self.__nBaud,
                code=self.__code)

        try:
            self.dataOut.systemHeaderObj = SystemHeader(self.__systemHeader)
        except:
            self.dataOut.systemHeaderObj = SystemHeader(nSamples=self.__nSamples,
                                                        nProfiles=nProfiles,
                                                        nChannels=len(
                                                            self.__channelList),
                                                        adcResolution=14)
        self.dataOut.type  = "Voltage"

        self.dataOut.data  = None

        self.dataOut.dtype = self.dtype

        # self.dataOut.nChannels = 0

        # self.dataOut.nHeights = 0

        self.dataOut.nProfiles   = int(nProfiles)

        self.dataOut.heightList  = self.__firstHeigth + \
            numpy.arange(self.__nSamples, dtype=numpy.float) * \
            self.__deltaHeigth

        self.dataOut.channelList = list(range(self.__num_subchannels))

        self.dataOut.blocksize   = self.dataOut.getNChannels() * self.dataOut.getNHeights()

        # self.dataOut.channelIndexList = None

        self.dataOut.flagNoData      = True

        self.dataOut.flagDataAsBlock = False
        # Set to TRUE if the data is discontinuous
        self.dataOut.flagDiscontinuousBlock = False

        self.dataOut.utctime     = None

        # timezone like jroheader, difference in minutes between UTC and localtime
        self.dataOut.timeZone    = self.__timezone / 60

        self.dataOut.dstFlag     = 0

        self.dataOut.errorCount  = 0

        try:
            self.dataOut.nCohInt = self.fixed_metadata_dict.get(
                'nCohInt', self.nCohInt)

            # asumo que la data esta decodificada
            self.dataOut.flagDecodeData = self.fixed_metadata_dict.get(
                'flagDecodeData', self.flagDecodeData)

            # asumo que la data esta sin flip
            self.dataOut.flagDeflipData = self.fixed_metadata_dict['flagDeflipData']

            self.dataOut.flagShiftFFT   = self.fixed_metadata_dict['flagShiftFFT']

            self.dataOut.useLocalTime   = self.fixed_metadata_dict['useLocalTime']
        except:
            pass

        self.dataOut.ippSeconds = ippSeconds

        # Time interval between profiles
        # self.dataOut.timeInterval = self.dataOut.ippSeconds * self.dataOut.nCohInt

        self.dataOut.frequency  = self.__frequency

        self.dataOut.realtime   = self.__online

    def findDatafiles(self, path, startDate=None, endDate=None):

        if not os.path.isdir(path):
            return []

        try:
            digitalReadObj = digital_rf.DigitalRFReader(
                path, load_all_metadata=True)
        except:
            digitalReadObj = digital_rf.DigitalRFReader(path)

        channelNameList    = digitalReadObj.get_channels()

        if not channelNameList:
            return []

        metadata_dict      = digitalReadObj.get_rf_file_metadata(channelNameList[0])

        sample_rate        = metadata_dict['sample_rate'][0]

        this_metadata_file = digitalReadObj.get_metadata(channelNameList[0])

        try:
            timezone       = this_metadata_file['timezone'].value
        except:
            timezone       = 0

        startUTCSecond, endUTCSecond = digitalReadObj.get_bounds(
            channelNameList[0]) / sample_rate - timezone

        startDatetime      = datetime.datetime.utcfromtimestamp(startUTCSecond)
        endDatatime        = datetime.datetime.utcfromtimestamp(endUTCSecond)

        if not startDate:
            startDate      = startDatetime.date()

        if not endDate:
            endDate        = endDatatime.date()

        dateList           = []

        thisDatetime       = startDatetime

        while(thisDatetime <= endDatatime):

            thisDate        = thisDatetime.date()

            if thisDate     < startDate:
                continue

            if thisDate     > endDate:
                break

            dateList.append(thisDate)
            thisDatetime += datetime.timedelta(1)

        return dateList

    def setup(self, path=None,
              startDate=None,
              endDate=None,
              startTime=datetime.time(0, 0, 0),
              endTime=datetime.time(23, 59, 59),
              channelList=None,
              nSamples=None,
              online=False,
              delay=60,
              buffer_size=1024,
              ippKm=None,
              nCohInt=1,
              nCode=1,
              nBaud=1,
              flagDecodeData=False,
              code=numpy.ones((1, 1), dtype=numpy.int),
              **kwargs):
        '''
        In this method we should set all initial parameters.

        Inputs:
            path
            startDate
            endDate
            startTime
            endTime
            set
            expLabel
            ext
            online
            delay
        '''
        self.path           = path
        self.nCohInt        = nCohInt
        self.flagDecodeData = flagDecodeData
        self.i              = 0
        if not os.path.isdir(path):
            raise ValueError("[Reading] Directory %s does not exist" % path)

        try:
            self.digitalReadObj = digital_rf.DigitalRFReader(
                path, load_all_metadata=True)
        except:
            self.digitalReadObj = digital_rf.DigitalRFReader(path)

        channelNameList         = self.digitalReadObj.get_channels()

        if not channelNameList:
            raise ValueError("[Reading] Directory %s does not have any files" % path)

        if not channelList:
            channelList = list(range(len(channelNameList)))

        ##########  Reading metadata ######################

        top_properties           = self.digitalReadObj.get_properties(
            channelNameList[channelList[0]])

        self.__num_subchannels   = top_properties['num_subchannels']
        self.__sample_rate       = 1.0 * \
            top_properties['sample_rate_numerator'] / \
            top_properties['sample_rate_denominator']
        # self.__samples_per_file = top_properties['samples_per_file'][0]
        self.__deltaHeigth       = 1e6 * 0.15 / self.__sample_rate  # why 0.15?

        this_metadata_file       = self.digitalReadObj.get_digital_metadata(
            channelNameList[channelList[0]])
        metadata_bounds          = this_metadata_file.get_bounds()
        self.fixed_metadata_dict = this_metadata_file.read(
            metadata_bounds[0])[metadata_bounds[0]]  # GET FIRST HEADER

        try:
            self.__processingHeader      = self.fixed_metadata_dict['processingHeader']
            self.__radarControllerHeader = self.fixed_metadata_dict['radarControllerHeader']
            self.__systemHeader          = self.fixed_metadata_dict['systemHeader']
            self.dtype                   = pickle.loads(self.fixed_metadata_dict['dtype'])
        except:
            pass

        self.__frequency = None

        self.__frequency = self.fixed_metadata_dict.get('frequency', 1)

        self.__timezone = self.fixed_metadata_dict.get('timezone', 18000)

        try:
            nSamples = self.fixed_metadata_dict['nSamples']
        except:
            nSamples = None

        self.__firstHeigth = 0

        try:
            codeType       = self.__radarControllerHeader['codeType']
        except:
            codeType       = 0

        try:
            if codeType:
                nCode = self.__radarControllerHeader['nCode']
                nBaud = self.__radarControllerHeader['nBaud']
                code  = self.__radarControllerHeader['code']
        except:
            pass

        if not ippKm:
            try:
                # seconds to km
                ippKm = self.__radarControllerHeader['ipp']
            except:
                ippKm = None
        ####################################################
        self.__ippKm   = ippKm
        startUTCSecond = None
        endUTCSecond   = None

        if startDate:
            startDatetime  = datetime.datetime.combine(startDate, startTime)
            startUTCSecond = (
                startDatetime - datetime.datetime(1970, 1, 1)).total_seconds() + self.__timezone

        if endDate:
            endDatetime   = datetime.datetime.combine(endDate, endTime)
            endUTCSecond  = (endDatetime - datetime.datetime(1970,
                                                            1, 1)).total_seconds() + self.__timezone

        start_index, end_index = self.digitalReadObj.get_bounds(
            channelNameList[channelList[0]])

        if not startUTCSecond:
            startUTCSecond = start_index / self.__sample_rate

        if start_index     > startUTCSecond * self.__sample_rate:
            startUTCSecond = start_index / self.__sample_rate

        if not endUTCSecond:
            endUTCSecond   = end_index / self.__sample_rate

        if end_index       < endUTCSecond * self.__sample_rate:
            endUTCSecond   = end_index / self.__sample_rate
        if not nSamples:
            if not ippKm:
                raise ValueError("[Reading] nSamples or ippKm should be defined")
            nSamples            = int(ippKm / (1e6 * 0.15 / self.__sample_rate))
        channelBoundList        = []
        channelNameListFiltered = []

        for thisIndexChannel in channelList:
            thisChannelName        = channelNameList[thisIndexChannel]
            start_index, end_index = self.digitalReadObj.get_bounds(
                thisChannelName)
            channelBoundList.append((start_index, end_index))
            channelNameListFiltered.append(thisChannelName)

        self.profileIndex = 0
        self.i            = 0
        self.__delay      = delay

        self.__codeType   = codeType
        self.__nCode      = nCode
        self.__nBaud      = nBaud
        self.__code       = code

        self.__datapath         = path
        self.__online           = online
        self.__channelList      = channelList
        self.__channelNameList  = channelNameListFiltered
        self.__channelBoundList = channelBoundList
        self.__nSamples         = nSamples
        self.__samples_to_read  = int(nSamples)  # FIJO: AHORA 40
        self.__nChannels        = len(self.__channelList)

        self.__startUTCSecond   = startUTCSecond
        self.__endUTCSecond     = endUTCSecond

        self.__timeInterval     = 1.0 * self.__samples_to_read / \
            self.__sample_rate  # Time interval

        if online:
            # self.__thisUnixSample = int(endUTCSecond*self.__sample_rate - 4*self.__samples_to_read)
            startUTCSecond = numpy.floor(endUTCSecond)

        # por que en el otro metodo lo primero q se hace es sumar samplestoread
        self.__thisUnixSample = int(startUTCSecond * self.__sample_rate) - self.__samples_to_read

        self.__data_buffer    = numpy.zeros(
            (self.__num_subchannels, self.__samples_to_read), dtype=numpy.complex)

        self.__setFileHeader()
        self.isConfig = True

        print("[Reading] Digital RF Data was found from %s to %s " % (
            datetime.datetime.utcfromtimestamp(
                self.__startUTCSecond - self.__timezone),
            datetime.datetime.utcfromtimestamp(
                self.__endUTCSecond - self.__timezone)
        ))

        print("[Reading] Starting process from %s to %s" % (datetime.datetime.utcfromtimestamp(startUTCSecond - self.__timezone),
                                                            datetime.datetime.utcfromtimestamp(
            endUTCSecond - self.__timezone)
        ))
        self.oldAverage    = None
        self.count         = 0
        self.executionTime = 0

    def __reload(self):
        #         print
        #         print "%s not in range [%s, %s]" %(
        #                                           datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone),
        #                                           datetime.datetime.utcfromtimestamp(self.__startUTCSecond - self.__timezone),
        #                                           datetime.datetime.utcfromtimestamp(self.__endUTCSecond - self.__timezone)
        #                                           )
        print("[Reading] reloading metadata ...")

        try:
            self.digitalReadObj.reload(complete_update=True)
        except:
            self.digitalReadObj = digital_rf.DigitalRFReader(self.path) 

        start_index, end_index  = self.digitalReadObj.get_bounds(
            self.__channelNameList[self.__channelList[0]])

        if start_index          > self.__startUTCSecond * self.__sample_rate:
            self.__startUTCSecond = 1.0 * start_index / self.__sample_rate

        if end_index            > self.__endUTCSecond * self.__sample_rate:
            self.__endUTCSecond = 1.0 * end_index / self.__sample_rate
            print()
            print("[Reading] New timerange found [%s, %s] " % (
                datetime.datetime.utcfromtimestamp(
                    self.__startUTCSecond - self.__timezone),
                datetime.datetime.utcfromtimestamp(
                    self.__endUTCSecond - self.__timezone)
            ))

            return True

        return False

    def timeit(self, toExecute):
        t0                  = time.time()
        toExecute()
        self.executionTime  = time.time() - t0
        if self.oldAverage is None:
            self.oldAverage = self.executionTime
        self.oldAverage     = (self.executionTime + self.count *
                           self.oldAverage) / (self.count + 1.0)
        self.count          = self.count + 1.0
        return

    def __readNextBlock(self, seconds=30, volt_scale=1):
        '''
        '''

        # Set the next data
        self.__flagDiscontinuousBlock = False
        self.__thisUnixSample        += self.__samples_to_read
        
        if self.__thisUnixSample + 2 * self.__samples_to_read > self.__endUTCSecond * self.__sample_rate:
            print ("[Reading] There are no more data into selected time-range")
            if self.__online:
                sleep(3)
                self.__reload()
            else:
                return False

            if self.__thisUnixSample + 2 * self.__samples_to_read > self.__endUTCSecond * self.__sample_rate:
                return False
                self.__thisUnixSample -= self.__samples_to_read

        indexChannel = 0

        dataOk = False
        
        for thisChannelName in self.__channelNameList:  # TODO VARIOS CHANNELS?
            for indexSubchannel in range(self.__num_subchannels):
                try:
                    t0     = time()
                    result = self.digitalReadObj.read_vector_c81d(self.__thisUnixSample,
                                                                  self.__samples_to_read,
                                                                  thisChannelName, sub_channel=indexSubchannel)
                    self.executionTime  = time() - t0
                    if self.oldAverage is None:
                        self.oldAverage = self.executionTime
                    self.oldAverage     = (
                        self.executionTime + self.count * self.oldAverage) / (self.count + 1.0)
                    self.count = self.count + 1.0

                except IOError as e:
                    # read next profile
                    self.__flagDiscontinuousBlock = True
                    print("[Reading] %s" % datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone), e)
                    break

                if result.shape[0] != self.__samples_to_read:
                    self.__flagDiscontinuousBlock = True
                    print("[Reading] %s: Too few samples were found, just %d/%d  samples" % (datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone),
                                                                                             result.shape[0],
                                                                                             self.__samples_to_read))
                    break
                
                self.__data_buffer[indexSubchannel, :] = result * volt_scale
                indexChannel+=1

                dataOk       = True

        self.__utctime       = self.__thisUnixSample / self.__sample_rate

        if not dataOk:
            return False

        print("[Reading] %s: %d samples <> %f sec" % (datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone),
                                                      self.__samples_to_read,
                                                      self.__timeInterval))

        self.__bufferIndex   = 0

        return True

    def __isBufferEmpty(self):
        return self.__bufferIndex > self.__samples_to_read - self.__nSamples  # 40960 - 40

    def getData(self, seconds=30, nTries=5):
        '''
            This method gets the data from files and put the data into the dataOut object

            In addition, increase el the buffer counter in one.

            Return:
                data    :    retorna un perfil de voltages (alturas * canales) copiados desde el
                            buffer. Si no hay mas archivos a leer retorna None.

            Affected:
                self.dataOut
                self.profileIndex
                self.flagDiscontinuousBlock
                self.flagIsNewBlock
        '''
        #print("getdata")
        err_counter = 0
        self.dataOut.flagNoData = True

        if self.__isBufferEmpty():
            #print("hi")
            self.__flagDiscontinuousBlock = False

            while True:
                #print ("q ha pasado")
                if self.__readNextBlock():
                    break
                if self.__thisUnixSample > self.__endUTCSecond * self.__sample_rate:
                    raise schainpy.admin.SchainError('Error')
                    return

                if self.__flagDiscontinuousBlock:
                    raise schainpy.admin.SchainError('discontinuous block found')
                    return

                if not self.__online:
                    raise schainpy.admin.SchainError('Online?')
                    return

                err_counter += 1
                if err_counter > nTries:
                    raise schainpy.admin.SchainError('Max retrys reach')
                    return

                print('[Reading] waiting %d seconds to read a new block' % seconds)
                time.sleep(seconds)

        self.dataOut.data                   = self.__data_buffer[:, self.__bufferIndex:self.__bufferIndex + self.__nSamples]
        self.dataOut.utctime                = ( self.__thisUnixSample + self.__bufferIndex) / self.__sample_rate
        self.dataOut.flagNoData             = False
        self.dataOut.flagDiscontinuousBlock = self.__flagDiscontinuousBlock
        self.dataOut.profileIndex           = self.profileIndex

        self.__bufferIndex += self.__nSamples
        self.profileIndex  += 1

        if self.profileIndex == self.dataOut.nProfiles:
            self.profileIndex = 0

        return True

    def printInfo(self):
        '''
        '''
        if self.__printInfo == False:
            return

        # self.systemHeaderObj.printInfo()
        # self.radarControllerHeaderObj.printInfo()

        self.__printInfo = False

    def printNumberOfBlock(self):
        '''
        '''
        return
        # print self.profileIndex

    def run(self, **kwargs):
        '''
        This method will be called many times so here you should put all your code
        '''
        
        if not self.isConfig:
            self.setup(**kwargs)
        #self.i = self.i+1
        self.getData(seconds=self.__delay)
        
        return

@MPDecorator
class DigitalRFWriter(Operation):
    '''
    classdocs
    '''

    def __init__(self, **kwargs):
        '''
        Constructor
        '''
        Operation.__init__(self, **kwargs)
        self.metadata_dict = {}
        self.dataOut       = None
        self.dtype         = None
        self.oldAverage    = 0

    def setHeader(self):

        self.metadata_dict['frequency']      = self.dataOut.frequency
        self.metadata_dict['timezone']       = self.dataOut.timeZone
        self.metadata_dict['dtype']          = pickle.dumps(self.dataOut.dtype)
        self.metadata_dict['nProfiles']      = self.dataOut.nProfiles
        self.metadata_dict['heightList']     = self.dataOut.heightList
        self.metadata_dict['channelList']    = self.dataOut.channelList
        self.metadata_dict['flagDecodeData'] = self.dataOut.flagDecodeData
        self.metadata_dict['flagDeflipData'] = self.dataOut.flagDeflipData
        self.metadata_dict['flagShiftFFT']   = self.dataOut.flagShiftFFT
        self.metadata_dict['useLocalTime']   = self.dataOut.useLocalTime
        self.metadata_dict['nCohInt']        = self.dataOut.nCohInt
        self.metadata_dict['type']           = self.dataOut.type
        self.metadata_dict['flagDataAsBlock']= getattr(
            self.dataOut, 'flagDataAsBlock', None)  # chequear

    def setup(self, dataOut, path, frequency, fileCadence, dirCadence, metadataCadence, set=0, metadataFile='metadata', ext='.h5'):
        '''
        In this method we should set all initial parameters.
        Input:
            dataOut: Input data will also be outputa data
        '''
        self.setHeader()
        self.__ippSeconds  = dataOut.ippSeconds
        self.__deltaH      = dataOut.getDeltaH()
        self.__sample_rate = 1e6 * 0.15 / self.__deltaH
        self.__dtype       = dataOut.dtype
        if len(dataOut.dtype) == 2:
            self.__dtype = dataOut.dtype[0]
        self.__nSamples  = dataOut.systemHeaderObj.nSamples
        self.__nProfiles = dataOut.nProfiles

        if self.dataOut.type != 'Voltage':
            raise 'Digital RF cannot be used with this data type'
            self.arr_data = numpy.ones((1, dataOut.nFFTPoints * len(
                self.dataOut.channelList)), dtype=[('r', self.__dtype), ('i', self.__dtype)])
        else:
            self.arr_data = numpy.ones((self.__nSamples, len(
                self.dataOut.channelList)), dtype=[('r', self.__dtype), ('i', self.__dtype)])

        file_cadence_millisecs  = 1000

        sample_rate_fraction    = Fraction(self.__sample_rate).limit_denominator()
        sample_rate_numerator   = int(sample_rate_fraction.numerator)
        sample_rate_denominator = int(sample_rate_fraction.denominator)
        start_global_index      = dataOut.utctime * self.__sample_rate

        uuid              = 'prueba'
        compression_level = 0
        checksum          = False
        is_complex        = True
        num_subchannels   = len(dataOut.channelList)
        is_continuous     = True
        marching_periods  = False

        self.digitalWriteObj = digital_rf.DigitalRFWriter(path, self.__dtype, dirCadence,
                                                          fileCadence, start_global_index,
                                                          sample_rate_numerator, sample_rate_denominator, uuid, compression_level, checksum,
                                                          is_complex, num_subchannels, is_continuous, marching_periods)
        metadata_dir         = os.path.join(path, 'metadata')
        os.system('mkdir %s' % (metadata_dir))
        self.digitalMetadataWriteObj = digital_rf.DigitalMetadataWriter(metadata_dir, dirCadence, 1,  # 236, file_cadence_millisecs / 1000
                                                                        sample_rate_numerator, sample_rate_denominator,
                                                                        metadataFile)
        self.isConfig      = True
        self.currentSample = 0
        self.oldAverage    = 0
        self.count         = 0
        return

    def writeMetadata(self):
        start_idx                                   = self.__sample_rate * self.dataOut.utctime

        self.metadata_dict['processingHeader']      = self.dataOut.processingHeaderObj.getAsDict(
        )
        self.metadata_dict['radarControllerHeader'] = self.dataOut.radarControllerHeaderObj.getAsDict(
        )
        self.metadata_dict['systemHeader']          = self.dataOut.systemHeaderObj.getAsDict(
        )
        self.digitalMetadataWriteObj.write(start_idx, self.metadata_dict)
        return

    def timeit(self, toExecute):
        t0 = time()
        toExecute()
        self.executionTime  = time() - t0
        if self.oldAverage is None:
            self.oldAverage = self.executionTime
        self.oldAverage     = (self.executionTime + self.count *
                           self.oldAverage) / (self.count + 1.0)
        self.count          = self.count + 1.0
        return

    def writeData(self):
        if self.dataOut.type != 'Voltage':
            raise 'Digital RF cannot be used with this data type'
            for channel in self.dataOut.channelList:
                for i in range(self.dataOut.nFFTPoints):
                    self.arr_data[1][channel * self.dataOut.nFFTPoints +
                                     i]['r'] = self.dataOut.data[channel][i].real
                    self.arr_data[1][channel * self.dataOut.nFFTPoints +
                                     i]['i'] = self.dataOut.data[channel][i].imag
        else:
            for i in range(self.dataOut.systemHeaderObj.nSamples):
                for channel in self.dataOut.channelList:
                    self.arr_data[i][channel]['r'] = self.dataOut.data[channel][i].real
                    self.arr_data[i][channel]['i'] = self.dataOut.data[channel][i].imag

        def f(): return self.digitalWriteObj.rf_write(self.arr_data)
        self.timeit(f)

        return

    def run(self, dataOut, frequency=49.92e6, path=None, fileCadence=1000, dirCadence=36000, metadataCadence=1, **kwargs):
        '''
        This method will be called many times so here you should put all your code
        Inputs:
            dataOut: object with the data
        '''
        # print dataOut.__dict__
        self.dataOut = dataOut
        if not self.isConfig:
            self.setup(dataOut, path, frequency, fileCadence,
                       dirCadence, metadataCadence, **kwargs)
            self.writeMetadata()

        self.writeData()

        ## self.currentSample += 1
        # if self.dataOut.flagDataAsBlock or self.currentSample == 1:
        # self.writeMetadata()
        ## if self.currentSample == self.__nProfiles: self.currentSample = 0

        return dataOut# en la version 2.7 no aparece este return

    def close(self):
        print('[Writing] - Closing files ')
        print('Average of writing to digital rf format is ', self.oldAverage * 1000)
        try:
            self.digitalWriteObj.close()
        except:
            pass



'''
Created on Jul 3, 2014

@author: roj-idl71
'''

import os, sys
import time, datetime
import numpy
import fnmatch
import glob
from time import sleep

try:
    import pyfits
except ImportError as e:
    pass

from xml.etree.ElementTree import ElementTree

from .jroIO_base import isRadarFolder, isNumber
from schainpy.model.data.jrodata import Fits
from schainpy.model.proc.jroproc_base import Operation, ProcessingUnit, MPDecorator
from schainpy.utils import log


class PyFits(object):
    name=None
    format=None
    array =None
    data =None
    thdulist=None
    prihdr=None
    hdu=None

    def __init__(self):

        pass

    def setColF(self,name,format,array):
        self.name=name
        self.format=format
        self.array=array
        a1=numpy.array([self.array],dtype=numpy.float32)
        self.col1 = pyfits.Column(name=self.name, format=self.format, array=a1)
        return self.col1

#    def setColP(self,name,format,data):
#        self.name=name
#        self.format=format
#        self.data=data
#        a2=numpy.array([self.data],dtype=numpy.float32)
#        self.col2 = pyfits.Column(name=self.name, format=self.format, array=a2)
#        return self.col2


    def writeData(self,name,format,data):
        self.name=name
        self.format=format
        self.data=data
        a2=numpy.array([self.data],dtype=numpy.float32)
        self.col2 = pyfits.Column(name=self.name, format=self.format, array=a2)
        return self.col2

    def cFImage(self,idblock,year,month,day,hour,minute,second):
        self.hdu= pyfits.PrimaryHDU(idblock)
        self.hdu.header.set("Year",year)
        self.hdu.header.set("Month",month)
        self.hdu.header.set("Day",day)
        self.hdu.header.set("Hour",hour)
        self.hdu.header.set("Minute",minute)
        self.hdu.header.set("Second",second)
        return self.hdu


    def Ctable(self,colList):
        self.cols=pyfits.ColDefs(colList)
        self.tbhdu = pyfits.new_table(self.cols)
        return self.tbhdu


    def CFile(self,hdu,tbhdu):
        self.thdulist=pyfits.HDUList([hdu,tbhdu])

    def wFile(self,filename):
        if os.path.isfile(filename):
            os.remove(filename)
        self.thdulist.writeto(filename)


class ParameterConf:
    ELEMENTNAME = 'Parameter'
    def __init__(self):
        self.name = ''
        self.value = ''

    def readXml(self, parmElement):
        self.name = parmElement.get('name')
        self.value = parmElement.get('value')

    def getElementName(self):
        return self.ELEMENTNAME

class Metadata(object):

    def __init__(self, filename):
        self.parmConfObjList = []
        self.readXml(filename)

    def readXml(self, filename):
        self.projectElement = None
        self.procUnitConfObjDict = {}
        self.projectElement = ElementTree().parse(filename)
        self.project = self.projectElement.tag

        parmElementList = self.projectElement.getiterator(ParameterConf().getElementName())

        for parmElement in parmElementList:
            parmConfObj = ParameterConf()
            parmConfObj.readXml(parmElement)
            self.parmConfObjList.append(parmConfObj)

@MPDecorator
class FitsWriter(Operation):
    def __init__(self, **kwargs):
        Operation.__init__(self, **kwargs)
        self.isConfig = False
        self.dataBlocksPerFile = None
        self.blockIndex = 0
        self.flagIsNewFile = 1
        self.fitsObj = None
        self.optchar = 'P'
        self.ext = '.fits'
        self.setFile = 0

    def setFitsHeader(self, dataOut, metadatafile=None):

        header_data = pyfits.PrimaryHDU()

        header_data.header['EXPNAME'] = "RADAR DATA"
        header_data.header['DATATYPE'] = "SPECTRA"
        header_data.header['COMMENT'] = ""

        if metadatafile:

            metadata4fits = Metadata(metadatafile)

            for parameter in metadata4fits.parmConfObjList:
                parm_name = parameter.name
                parm_value = parameter.value

                header_data.header[parm_name] = parm_value

        header_data.header['DATETIME'] = time.strftime("%b %d %Y %H:%M:%S", dataOut.datatime.timetuple())
        header_data.header['CHANNELLIST'] = str(dataOut.channelList)
        header_data.header['NCHANNELS'] = dataOut.nChannels
        #header_data.header['HEIGHTS'] = dataOut.heightList
        header_data.header['NHEIGHTS'] = dataOut.nHeights

        header_data.header['IPPSECONDS'] = dataOut.ippSeconds
        header_data.header['NCOHINT'] = dataOut.nCohInt
        header_data.header['NINCOHINT'] = dataOut.nIncohInt
        header_data.header['TIMEZONE'] = dataOut.timeZone
        header_data.header['NBLOCK'] = self.blockIndex

        header_data.writeto(self.filename)

        self.addExtension(dataOut.heightList,'HEIGHTLIST')


    def setup(self, dataOut, path, dataBlocksPerFile=100, metadatafile=None):

        self.path = path
        self.dataOut = dataOut
        self.metadatafile = metadatafile
        self.dataBlocksPerFile = dataBlocksPerFile

    def open(self):
        self.fitsObj = pyfits.open(self.filename, mode='update')


    def addExtension(self, data, tagname):
        self.open()
        extension = pyfits.ImageHDU(data=data, name=tagname)
        #extension.header['TAG'] = tagname
        self.fitsObj.append(extension)
        self.write()

    def addData(self, data):
        self.open()
        extension = pyfits.ImageHDU(data=data, name=self.fitsObj[0].header['DATATYPE'])
        extension.header['UTCTIME'] = self.dataOut.utctime
        self.fitsObj.append(extension)
        self.blockIndex += 1
        self.fitsObj[0].header['NBLOCK'] = self.blockIndex

        self.write()

    def write(self):

        self.fitsObj.flush(verbose=True)
        self.fitsObj.close()


    def setNextFile(self):

        ext = self.ext
        path = self.path

        timeTuple = time.localtime( self.dataOut.utctime)
        subfolder = 'd%4.4d%3.3d' % (timeTuple.tm_year,timeTuple.tm_yday)

        fullpath = os.path.join( path, subfolder )
        if not( os.path.exists(fullpath) ):
            os.mkdir(fullpath)
            self.setFile = -1 #inicializo mi contador de seteo
        else:
            filesList = os.listdir( fullpath )
            if len( filesList ) > 0:
                filesList = sorted( filesList, key=str.lower )
                filen = filesList[-1]

                if isNumber( filen[8:11] ):
                    self.setFile = int( filen[8:11] ) #inicializo mi contador de seteo al seteo del ultimo file
                else:
                    self.setFile = -1
            else:
                self.setFile = -1 #inicializo mi contador de seteo

        setFile = self.setFile
        setFile += 1

        thisFile = '%s%4.4d%3.3d%3.3d%s' % (self.optchar,
                                        timeTuple.tm_year,
                                        timeTuple.tm_yday,
                                        setFile,
                                        ext )

        filename = os.path.join( path, subfolder, thisFile )

        self.blockIndex = 0
        self.filename = filename
        self.setFile = setFile
        self.flagIsNewFile = 1

        print('Writing the file: %s'%self.filename)

        self.setFitsHeader(self.dataOut, self.metadatafile)

        return 1

    def writeBlock(self):
        self.addData(self.dataOut.data_spc)
        self.flagIsNewFile = 0


    def __setNewBlock(self):

        if self.flagIsNewFile:
            return 1

        if self.blockIndex < self.dataBlocksPerFile:
            return 1

        if not( self.setNextFile() ):
            return 0

        return 1

    def writeNextBlock(self):
        if not( self.__setNewBlock() ):
            return 0
        self.writeBlock()
        return 1

    def putData(self):
        if self.flagIsNewFile:
            self.setNextFile()
        self.writeNextBlock()

    def run(self, dataOut, path, dataBlocksPerFile=100, metadatafile=None, **kwargs):
        if not(self.isConfig):
            self.setup(dataOut, path, dataBlocksPerFile=dataBlocksPerFile, metadatafile=metadatafile, **kwargs)
            self.isConfig = True
        self.putData()


class FitsReader(ProcessingUnit):

#     __TIMEZONE = time.timezone

    expName = None
    datetimestr = None
    utc = None
    nChannels = None
    nSamples = None
    dataBlocksPerFile = None
    comments = None
    lastUTTime = None
    header_dict = None
    data = None
    data_header_dict = None

    def __init__(self):#, **kwargs):
        ProcessingUnit.__init__(self)#, **kwargs)
        self.isConfig = False
        self.ext = '.fits'
        self.setFile = 0
        self.flagNoMoreFiles = 0
        self.flagIsNewFile = 1
        self.flagDiscontinuousBlock = None
        self.fileIndex = None
        self.filename = None
        self.fileSize = None
        self.fitsObj = None
        self.timeZone = None
        self.nReadBlocks = 0
        self.nTotalBlocks = 0
        self.dataOut = self.createObjByDefault()
        self.maxTimeStep = 10# deberia ser definido por el usuario usando el metodo setup()
        self.blockIndex = 1

    def createObjByDefault(self):

        dataObj = Fits()

        return dataObj

    def isFileinThisTime(self, filename, startTime, endTime, useLocalTime=False):
        try:
            fitsObj = pyfits.open(filename,'readonly')
        except:
            print("File %s can't be opened" %(filename))
            return None

        header = fitsObj[0].header
        struct_time = time.strptime(header['DATETIME'], "%b %d %Y %H:%M:%S")
        utc = time.mktime(struct_time) - time.timezone #TIMEZONE debe ser un parametro del header FITS

        ltc = utc
        if useLocalTime:
            ltc -= time.timezone
        thisDatetime = datetime.datetime.utcfromtimestamp(ltc)
        thisTime = thisDatetime.time()

        if not ((startTime <= thisTime) and (endTime > thisTime)):
            return None

        return thisDatetime

    def __setNextFileOnline(self):
        raise NotImplementedError

    def __setNextFileOffline(self):
        idFile = self.fileIndex

        while (True):
            idFile += 1
            if not(idFile < len(self.filenameList)):
                self.flagNoMoreFiles = 1
                print("No more Files")
                return 0

            filename = self.filenameList[idFile]

#            if not(self.__verifyFile(filename)):
#                continue

            fileSize = os.path.getsize(filename)
            fitsObj = pyfits.open(filename,'readonly')
            break

        self.flagIsNewFile = 1
        self.fileIndex = idFile
        self.filename = filename
        self.fileSize = fileSize
        self.fitsObj = fitsObj
        self.blockIndex = 0
        print("Setting the file: %s"%self.filename)

        return 1

    def __setValuesFromHeader(self):

        self.dataOut.header = self.header_dict
        self.dataOut.expName = self.expName

        self.dataOut.timeZone = self.timeZone
        self.dataOut.dataBlocksPerFile = self.dataBlocksPerFile
        self.dataOut.comments = self.comments
#         self.dataOut.timeInterval = self.timeInterval
        self.dataOut.channelList = self.channelList
        self.dataOut.heightList = self.heightList

        self.dataOut.nCohInt = self.nCohInt
        self.dataOut.nIncohInt = self.nIncohInt
        self.dataOut.ipp_sec = self.ippSeconds

    def readHeader(self):
        headerObj = self.fitsObj[0]

        self.header_dict = headerObj.header
        if 'EXPNAME' in list(headerObj.header.keys()):
            self.expName = headerObj.header['EXPNAME']

        if 'DATATYPE' in list(headerObj.header.keys()):
            self.dataType = headerObj.header['DATATYPE']

        self.datetimestr = headerObj.header['DATETIME']
        channelList = headerObj.header['CHANNELLIST']
        channelList = channelList.split('[')
        channelList = channelList[1].split(']')
        channelList = channelList[0].split(',')
        channelList = [int(ch) for ch in channelList]
        self.channelList = channelList
        self.nChannels = headerObj.header['NCHANNELS']
        self.nHeights = headerObj.header['NHEIGHTS']
        self.ippSeconds = headerObj.header['IPPSECONDS']
        self.nCohInt = headerObj.header['NCOHINT']
        self.nIncohInt = headerObj.header['NINCOHINT']
        self.dataBlocksPerFile = headerObj.header['NBLOCK']
        self.timeZone = headerObj.header['TIMEZONE']

#         self.timeInterval = self.ippSeconds * self.nCohInt * self.nIncohInt

        if 'COMMENT' in list(headerObj.header.keys()):
            self.comments = headerObj.header['COMMENT']

        self.readHeightList()

    def readHeightList(self):
        self.blockIndex = self.blockIndex + 1
        obj = self.fitsObj[self.blockIndex]
        self.heightList = obj.data
        self.blockIndex = self.blockIndex + 1

    def readExtension(self):
        obj = self.fitsObj[self.blockIndex]
        self.heightList = obj.data
        self.blockIndex = self.blockIndex + 1

    def setNextFile(self):

        if self.online:
            newFile = self.__setNextFileOnline()
        else:
            newFile = self.__setNextFileOffline()

        if not(newFile):
            return 0

        self.readHeader()
        self.__setValuesFromHeader()
        self.nReadBlocks = 0
#         self.blockIndex = 1
        return 1

    def searchFilesOffLine(self,
                            path,
                            startDate,
                            endDate,
                            startTime=datetime.time(0,0,0),
                            endTime=datetime.time(23,59,59),
                            set=None,
                            expLabel='',
                            ext='.fits',
                            walk=True):

        pathList = []

        if not walk:
            pathList.append(path)

        else:
            dirList = []
            for thisPath in os.listdir(path):
                if not os.path.isdir(os.path.join(path,thisPath)):
                    continue
                if not isRadarFolder(thisPath):
                    continue

                dirList.append(thisPath)

            if not(dirList):
                return None, None

            thisDate = startDate

            while(thisDate <= endDate):
                year = thisDate.timetuple().tm_year
                doy = thisDate.timetuple().tm_yday

                matchlist = fnmatch.filter(dirList, '?' + '%4.4d%3.3d' % (year,doy) + '*')
                if len(matchlist) == 0:
                    thisDate += datetime.timedelta(1)
                    continue
                for match in matchlist:
                    pathList.append(os.path.join(path,match,expLabel))

                thisDate += datetime.timedelta(1)

        if pathList == []:
            print("Any folder was found for the date range: %s-%s" %(startDate, endDate))
            return None, None

        print("%d folder(s) was(were) found for the date range: %s - %s" %(len(pathList), startDate, endDate))

        filenameList = []
        datetimeList = []

        for i in range(len(pathList)):

            thisPath = pathList[i]

            fileList = glob.glob1(thisPath, "*%s" %ext)
            fileList.sort()

            for thisFile in fileList:

                filename = os.path.join(thisPath,thisFile)
                thisDatetime = self.isFileinThisTime(filename, startTime, endTime)

                if not(thisDatetime):
                    continue

                filenameList.append(filename)
                datetimeList.append(thisDatetime)

        if not(filenameList):
            print("Any file was found for the time range %s - %s" %(startTime, endTime))
            return None, None

        print("%d file(s) was(were) found for the time range: %s - %s" %(len(filenameList), startTime, endTime))
        print()

        for i in range(len(filenameList)):
            print("%s -> [%s]" %(filenameList[i], datetimeList[i].ctime()))

        self.filenameList = filenameList
        self.datetimeList = datetimeList

        return pathList, filenameList

    def setup(self, path=None,
                startDate=None,
                endDate=None,
                startTime=datetime.time(0,0,0),
                endTime=datetime.time(23,59,59),
                set=0,
                expLabel = "",
                ext = None,
                online = False,
                delay = 60,
                walk = True):

        if path == None:
            raise ValueError("The path is not valid")

        if ext == None:
            ext = self.ext

        if not(online):
            print("Searching files in offline mode ...")
            pathList, filenameList = self.searchFilesOffLine(path, startDate=startDate, endDate=endDate,
                                                               startTime=startTime, endTime=endTime,
                                                               set=set, expLabel=expLabel, ext=ext,
                                                               walk=walk)

            if not(pathList):
                print("No *%s files into the folder %s \nfor the range: %s - %s"%(ext, path,
                                                        datetime.datetime.combine(startDate,startTime).ctime(),
                                                        datetime.datetime.combine(endDate,endTime).ctime()))

                sys.exit(-1)

            self.fileIndex = -1
            self.pathList = pathList
            self.filenameList = filenameList

        self.online = online
        self.delay = delay
        ext = ext.lower()
        self.ext = ext

        if not(self.setNextFile()):
            if (startDate!=None) and (endDate!=None):
                print("No files in range: %s - %s" %(datetime.datetime.combine(startDate,startTime).ctime(), datetime.datetime.combine(endDate,endTime).ctime()))
            elif startDate != None:
                print("No files in range: %s" %(datetime.datetime.combine(startDate,startTime).ctime()))
            else:
                print("No files")

            sys.exit(-1)



    def readBlock(self):
        dataObj = self.fitsObj[self.blockIndex]

        self.data = dataObj.data
        self.data_header_dict = dataObj.header
        self.utc = self.data_header_dict['UTCTIME']

        self.flagIsNewFile = 0
        self.blockIndex += 1
        self.nTotalBlocks += 1
        self.nReadBlocks += 1

        return 1

    def __jumpToLastBlock(self):
        raise NotImplementedError

    def __waitNewBlock(self):
        """
        Return 1 si se encontro un nuevo bloque de datos, 0 de otra forma.

        Si el modo de lectura es OffLine siempre retorn 0
        """
        if not self.online:
            return 0

        if (self.nReadBlocks >= self.dataBlocksPerFile):
            return 0

        currentPointer = self.fp.tell()

        neededSize = self.processingHeaderObj.blockSize + self.basicHeaderSize

        for nTries in range( self.nTries ):

            self.fp.close()
            self.fp = open( self.filename, 'rb' )
            self.fp.seek( currentPointer )

            self.fileSize = os.path.getsize( self.filename )
            currentSize = self.fileSize - currentPointer

            if ( currentSize >= neededSize ):
                self.__rdBasicHeader()
                return 1

            print("\tWaiting %0.2f seconds for the next block, try %03d ..." % (self.delay, nTries+1))
            sleep( self.delay )


        return 0

    def __setNewBlock(self):

        if self.online:
            self.__jumpToLastBlock()

        if self.flagIsNewFile:
            return 1

        self.lastUTTime = self.utc

        if self.online:
            if self.__waitNewBlock():
                return 1

        if self.nReadBlocks < self.dataBlocksPerFile:
            return 1

        if not(self.setNextFile()):
            return 0

        deltaTime = self.utc - self.lastUTTime

        self.flagDiscontinuousBlock = 0

        if deltaTime > self.maxTimeStep:
            self.flagDiscontinuousBlock = 1

        return 1


    def readNextBlock(self):
        if not(self.__setNewBlock()):
            return 0

        if not(self.readBlock()):
            return 0

        return 1

    def printInfo(self):

        pass

    def getData(self):

        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            return (0, 'No more files')

        self.flagDiscontinuousBlock = 0
        self.flagIsNewBlock = 0

        if not(self.readNextBlock()):
            return (1, 'Error reading data')

        if self.data is None:
            self.dataOut.flagNoData = True
            return (0, 'No more data')

        self.dataOut.data = self.data
        self.dataOut.data_header = self.data_header_dict
        self.dataOut.utctime = self.utc

#         self.dataOut.header = self.header_dict
#         self.dataOut.expName = self.expName
#         self.dataOut.nChannels = self.nChannels
#         self.dataOut.timeZone = self.timeZone
#         self.dataOut.dataBlocksPerFile = self.dataBlocksPerFile
#         self.dataOut.comments = self.comments
# #         self.dataOut.timeInterval = self.timeInterval
#         self.dataOut.channelList = self.channelList
#         self.dataOut.heightList = self.heightList
        self.dataOut.flagNoData = False
        # return self.dataOut.data

    def run(self, **kwargs):

        if not(self.isConfig):
            self.setup(**kwargs)
            self.isConfig = True

        self.getData()

@MPDecorator
class SpectraHeisWriter(Operation):
#    set = None
    setFile = None
    idblock = None
    doypath = None
    subfolder = None

    def __init__(self):#, **kwargs):
        Operation.__init__(self)#, **kwargs)
        self.wrObj = PyFits()
#        self.dataOut = dataOut
        self.nTotalBlocks=0
#        self.set = None
        self.setFile = None
        self.idblock = 0
        self.wrpath = None
        self.doypath = None
        self.subfolder = None
        self.isConfig = False

    def isNumber(str):
        """
        Chequea si el conjunto de caracteres que componen un string puede ser convertidos a un numero.

        Excepciones:
        Si un determinado string no puede ser convertido a numero
        Input:
        str, string al cual se le analiza para determinar si convertible a un numero o no

        Return:
        True    :    si el string es uno numerico
        False   :    no es un string numerico
        """
        try:
            float( str )
            return True
        except:
            return False

    def setup(self, dataOut, wrpath):

        if not(os.path.exists(wrpath)):
            os.mkdir(wrpath)

        self.wrpath = wrpath
#        self.setFile = 0
        self.dataOut = dataOut

    def putData(self):
        name= time.localtime( self.dataOut.utctime)
        ext=".fits"

        if self.doypath == None:
           self.subfolder = 'F%4.4d%3.3d_%d' % (name.tm_year,name.tm_yday,time.mktime(datetime.datetime.now().timetuple()))
           self.doypath = os.path.join( self.wrpath, self.subfolder )
           os.mkdir(self.doypath)

        if self.setFile == None:
#           self.set = self.dataOut.set
           self.setFile = 0
#        if self.set != self.dataOut.set:
##            self.set = self.dataOut.set
#            self.setFile = 0

        #make the filename
        thisFile = 'D%4.4d%3.3d_%3.3d%s' % (name.tm_year,name.tm_yday,self.setFile,ext)

        filename = os.path.join(self.wrpath,self.subfolder, thisFile)

        idblock = numpy.array([self.idblock],dtype="int64")
        header=self.wrObj.cFImage(idblock=idblock,
                                year=time.gmtime(self.dataOut.utctime).tm_year,
                                month=time.gmtime(self.dataOut.utctime).tm_mon,
                                day=time.gmtime(self.dataOut.utctime).tm_mday,
                                hour=time.gmtime(self.dataOut.utctime).tm_hour,
                                minute=time.gmtime(self.dataOut.utctime).tm_min,
                                second=time.gmtime(self.dataOut.utctime).tm_sec)

        c=3E8
        deltaHeight = self.dataOut.heightList[1] - self.dataOut.heightList[0]
        freq=numpy.arange(-1*self.dataOut.nHeights/2.,self.dataOut.nHeights/2.)*(c/(2*deltaHeight*1000))

        colList = []

        colFreq=self.wrObj.setColF(name="freq", format=str(self.dataOut.nFFTPoints)+'E', array=freq)

        colList.append(colFreq)

        nchannel=self.dataOut.nChannels

        for i in range(nchannel):
            col = self.wrObj.writeData(name="PCh"+str(i+1),
                                         format=str(self.dataOut.nFFTPoints)+'E',
                                          data=10*numpy.log10(self.dataOut.data_spc[i,:]))

            colList.append(col)

        data=self.wrObj.Ctable(colList=colList)

        self.wrObj.CFile(header,data)

        self.wrObj.wFile(filename)

        #update the setFile
        self.setFile += 1
        self.idblock += 1

        return 1

    def run(self, dataOut, **kwargs):

        if not(self.isConfig):

            self.setup(dataOut, **kwargs)
            self.isConfig = True

        self.putData()
        return dataOut


'''
@author: Daniel Suarez
'''
import numpy
from .jroproc_base import ProcessingUnit, Operation
from schainpy.model.data.jroamisr import AMISR

class AMISRProc(ProcessingUnit):
    def __init__(self, **kwargs):
        ProcessingUnit.__init__(self, **kwargs)
        self.objectDict = {}
        self.dataOut = AMISR()
        
    def run(self):
        if self.dataIn.type == 'AMISR':
            self.dataOut.copy(self.dataIn) 


class PrintInfoAMISR(Operation):
    def __init__(self, **kwargs):
        Operation.__init__(self, **kwargs)
        self.__isPrinted = False
    
    def run(self, dataOut):
        
        if not self.__isPrinted:
            print('Number of Records by File: %d'%dataOut.nRecords)
            print('Number of Pulses: %d'%dataOut.nProfiles)
            print('Number of Pulses by Frame: %d'%dataOut.npulseByFrame)
            print('Number of Samples by Pulse: %d'%len(dataOut.heightList))
            print('Ipp Seconds: %f'%dataOut.ippSeconds)
            print('Number of Beams: %d'%dataOut.nBeams)
            print('BeamCodes:')
            beamStrList = ['Beam %d -> Code=%d, azimuth=%2.2f,  zenith=%2.2f, gain=%2.2f'%(k,v[0],v[1],v[2],v[3]) for k,v in list(dataOut.beamCodeDict.items())]
            for b in beamStrList:
                print(b)
            self.__isPrinted = True
        
        return
        
        
class BeamSelector(Operation):
    profileIndex = None
    nProfiles = None
    
    def __init__(self, **kwargs):
        Operation.__init__(self, **kwargs)
        self.profileIndex = 0
        self.__isConfig = False
    
    def incIndex(self):
        self.profileIndex += 1
        
        if self.profileIndex >= self.nProfiles:
            self.profileIndex = 0
    
    def isProfileInRange(self, minIndex, maxIndex):
        
        if self.profileIndex < minIndex:
            return False
        
        if self.profileIndex > maxIndex:
            return False
        
        return True
    
    def isProfileInList(self, profileList):
        
        if self.profileIndex not in profileList:
            return False
        
        return True
    
    def run(self, dataOut, beam=None):
        
        dataOut.flagNoData = True
        
        if not(self.__isConfig):
            
            self.nProfiles = dataOut.nProfiles
            self.profileIndex = dataOut.profileIndex
            self.__isConfig = True

        if beam != None:
            if self.isProfileInList(dataOut.beamRangeDict[beam]):
                beamInfo = dataOut.beamCodeDict[beam]
                dataOut.azimuth = beamInfo[1]
                dataOut.zenith = beamInfo[2]
                dataOut.gain = beamInfo[3]
                dataOut.flagNoData = False
                
            self.incIndex()
            return 1
        
        else:
            raise ValueError("BeamSelector needs beam value")
        
        return 0

class ProfileToChannels(Operation):

    def __init__(self, **kwargs):
        Operation.__init__(self, **kwargs)
        self.__isConfig = False
        self.__counter_chan = 0
        self.buffer = None

    def isProfileInList(self, profileList):
        
        if self.profileIndex not in profileList:
            return False
        
        return True
    
    def run(self, dataOut):
        
        dataOut.flagNoData = True
        
        if not(self.__isConfig):
            nchannels = len(list(dataOut.beamRangeDict.keys()))
            nsamples = dataOut.nHeights
            self.buffer = numpy.zeros((nchannels, nsamples), dtype = 'complex128')
            dataOut.beam.codeList = [dataOut.beamCodeDict[x][0] for x in range(nchannels)]
            dataOut.beam.azimuthList = [dataOut.beamCodeDict[x][1] for x in range(nchannels)]
            dataOut.beam.zenithList = [dataOut.beamCodeDict[x][2] for x in range(nchannels)]
            self.__isConfig = True
        
        for i in range(self.buffer.shape[0]):
            if dataOut.profileIndex in dataOut.beamRangeDict[i]:
                self.buffer[i,:] = dataOut.data
                break
        
        
        self.__counter_chan += 1
          
        if self.__counter_chan >= self.buffer.shape[0]:
            self.__counter_chan = 0
            dataOut.data = self.buffer.copy()
            dataOut.channelList = list(range(self.buffer.shape[0]))
            self.__isConfig = False
            dataOut.flagNoData = False
        pass
              


voltage = '''import os, sys, time 
from schainpy.controller import Project


def main():
    desc = "{desc}"
    controller = Project()
    controller.setup(id='200', name="{name}", description=desc)

    read_unit = controller.addReadUnit(datatype='Voltage',
                                        path="{path}",
                                        startDate="{startDate}",
                                        endDate="{endDate}",
                                        startTime="{startHour}",
                                        endTime="{endHour}",
                                        online=0,
                                        verbose=1,
                                        walk=0,
                                        delay=180,
                                        )

    code = '[[1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1]]'
    nCode = '128'
    nBaud = '3'


    proc_voltage = controller.addProcUnit(name='VoltageProc', inputId=read_unit.getId())

    op1 = proc_voltage.addOperation(name='selectChannels', optype='self')
    op1.addParameter(name='channelList', value='0, 1, 2, 3', format='intlist')

    op2 = proc_voltage.addOperation(name='filterByHeights', optype='self')
    op2.addParameter(name='window', value='4', format='int')

    op3 = proc_voltage.addOperation(name='ProfileSelector', optype='other')
    op3.addParameter(name='profileRangeList', value='32, 159', format='intList')

    op4 = proc_voltage.addOperation(name='Decoder', optype='other')
    op4.addParameter(name='code', value=code, format='intlist')
    op4.addParameter(name='nCode', value=nCode, format='int')
    op4.addParameter(name='nBaud', value=nBaud, format='int')
    op4.addParameter(name='mode', value='0', format='int')
    
    op5 = proc_voltage.addOperation(name='Scope', optype='external')
    op5.addParameter(name='id', value='30', format='int')
    




    controller.start()

if __name__ == '__main__':
    import time
    start_time = time.time()
    main()
    print("--- %s seconds ---" % (time.time() - start_time))

'''


spectra = '''import os, sys, time 
from schainpy.controller import Project


def main():
    desc = "{desc}"
    controller = Project()
    controller.setup(id='300', name="{name}", description=desc)

    read_unit = controller.addReadUnit(datatype='Spectra',
                                        path="{path}",
                                        startDate="{startDate}",
                                        endDate="{endDate}",
                                        startTime="{startHour}",
                                        endTime="{endHour}",
                                        online=0,
                                        verbose=1,
                                        walk=0,
                                        delay=180,
                                        )

    proc_spectra = controller.addProcUnit(datatype='Spectra', inputId=read_unit.getId())
    proc_spectra.addParameter(name='nFFTPoints', value='128', format='int')  
    proc_spectra.addParameter(name='nProfiles', value='128', format='int')
    proc_spectra.addParameter(name='pairsList', value='(0, 1), (2, 3)', format='pairslist')

    op1 = proc_spectra.addOperation(name='IncohInt', optype='other')
    op1.addParameter(name='n', value='4', format='int')

    op2 = proc_spectra.addOperation(name='CrossSpectraPlot', optype='external')
    op2.addParameter(name='id', value='10', format='int')
    op2.addParameter(name='zmin', value='10.0', format='float')
    op2.addParameter(name='zmax', value='35.0', format='float')
        

    op3 = proc_spectra.addOperation(name='RTIPlot', optype='external')
    op3.addParameter(name='id', value='20', format='int')
    op3.addParameter(name='wintitle', value='RTI', format='str')
    op3.addParameter(name='xmin', value='0', format='float')
    op3.addParameter(name='xmax', value='24', format='float')
    op3.addParameter(name='zmin', value='12', format='int')
    op3.addParameter(name='zmax', value='32', format='int')
    op3.addParameter(name='showprofile', value='1', format='int')
    op3.addParameter(name='timerange', value=str(24*60*60), format='int')

    op4 = proc_spectra.addOperation(name='CoherenceMap', optype='external')
    op4.addParameter(name='id', value='30', format='int')
    op4.addParameter(name='xmin', value='0.0', format='float')
    op4.addParameter(name='xmax', value='24.0', format='float')


    controller.start()

if __name__ == '__main__':
    import time
    start_time = time.time()
    main()
    print("--- %s seconds ---" % (time.time() - start_time))

'''

voltagespectra = '''import os, sys, time 
from schainpy.controller import Project


def main():
    desc = "{desc}"
    controller = Project()
    controller.setup(id='400', name="{name}", description=desc)

    read_unit = controller.addReadUnit(datatype='Voltage',
                                        path="{path}",
                                        startDate="{startDate}",
                                        endDate="{endDate}",
                                        startTime="{startHour}",
                                        endTime="{endHour}",
                                        online=0,
                                        verbose=1,
                                        walk=0,
                                        delay=180,
                                        )

    code = '[[1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [1, 1, -1], [-1, -1, 1], [-1, -1, 1], [-1, -1, 1], [1, 1, -1]]'
    nCode = '128'
    nBaud = '3'


    proc_voltage = controller.addProcUnit(name='VoltageProc', inputId=read_unit.getId())

    op1 = proc_voltage.addOperation(name='selectChannels', optype='self')
    op1.addParameter(name='channelList', value='0, 1, 2, 3', format='intlist')

    op2 = proc_voltage.addOperation(name='filterByHeights', optype='self')
    op2.addParameter(name='window', value='4', format='int')

    op3 = proc_voltage.addOperation(name='ProfileSelector', optype='other')
    op3.addParameter(name='profileRangeList', value='32, 159', format='intList')

    op4 = proc_voltage.addOperation(name='Decoder', optype='other')
    op4.addParameter(name='code', value=code, format='intlist')
    op4.addParameter(name='nCode', value=nCode, format='int')
    op4.addParameter(name='nBaud', value=nBaud, format='int')
    op4.addParameter(name='mode', value='0', format='int')



    proc_spectra = controller.addProcUnit(datatype='Spectra', inputId=proc_voltage.getId())
    proc_spectra.addParameter(name='nFFTPoints', value='128', format='int')  
    proc_spectra.addParameter(name='nProfiles', value='128', format='int')
    proc_spectra.addParameter(name='pairsList', value='(0, 1), (2, 3)', format='pairslist')

    op5 = proc_spectra.addOperation(name='IncohInt', optype='other')
    op5.addParameter(name='n', value='4', format='int')

    op6 = proc_spectra.addOperation(name='CrossSpectraPlot', optype='external')
    op6.addParameter(name='id', value='10', format='int')
    op6.addParameter(name='zmin', value='10.0', format='float')
    op6.addParameter(name='zmax', value='35.0', format='float')
        

    op7 = proc_spectra.addOperation(name='RTIPlot', optype='external')
    op7.addParameter(name='id', value='20', format='int')
    op7.addParameter(name='wintitle', value='RTI', format='str')
    op7.addParameter(name='xmin', value='0', format='float')
    op7.addParameter(name='xmax', value='24', format='float')
    op7.addParameter(name='zmin', value='12', format='int')
    op7.addParameter(name='zmax', value='32', format='int')
    op7.addParameter(name='showprofile', value='1', format='int')
    op7.addParameter(name='timerange', value=str(24*60*60), format='int')

    op8 = proc_spectra.addOperation(name='CoherenceMap', optype='external')
    op8.addParameter(name='id', value='30', format='int')
    op8.addParameter(name='xmin', value='0.0', format='float')
    op8.addParameter(name='xmax', value='24.0', format='float')


    controller.start()

if __name__ == '__main__':
    import time
    start_time = time.time()
    main()
    print("--- %s seconds ---" % (time.time() - start_time))

'''








multiprocess = '''from schainpy.controller import Project, MPProject
from time import sleep
desc = "{desc}"

####################
# PLOTTER RECEIVER #
####################
plotter = Project()
plotter.setup(id='100', name='receiver', description=desc)

receiver_plot = plotter.addProcUnit(name='PlotterReceiver')
receiver_plot.addParameter(name='throttle', value=20, format='int')
receiver_plot.addParameter(name='plottypes', value='rti', format='str')

rti = receiver_plot.addOperation(name='PlotRTIData', optype='other')
rti.addParameter(name='zmin', value='-40.0', format='float') 
rti.addParameter(name='zmax', value='100.0', format='float') 
rti.addParameter(name='decimation', value='200', format='int') 
rti.addParameter(name='xmin', value='0.0', format='int') 
rti.addParameter(name='colormap', value='jet', format='str') 

plotter.start()

sleep(2)

################
# DATA EMITTER #
################
controller = Project()
controller.setup(id='200', name="{name}", description=desc)

spectra_reader = controller.addReadUnit(datatype='SpectraReader',
                                        path="{path}",
                                        startDate={startDate},
                                        endDate={endDate},
                                        startTime="{startHour}",
                                        endTime="{endHour}",
                                        online=0,
                                        verbose=1,
                                        walk=1,
                                        )

spectra_proc = controller.addProcUnit(datatype='Spectra', inputId=spectra_reader.getId())

parameters_proc = controller.addProcUnit(datatype='ParametersProc', inputId=spectra_proc.getId())
moments = parameters_proc.addOperation(name='SpectralMoments', optype='other')

publish = parameters_proc.addOperation(name='PublishData', optype='other')
publish.addParameter(name='zeromq', value=1, format='int')
publish.addParameter(name='verbose', value=0, format='bool')

MPProject(controller, 16)


'''



"""
Utilities for IO modules
"""

import os
from datetime import datetime

def folder_in_range(folder, start_date, end_date, pattern):
    """
    Check whether folder is bettwen start_date and end_date

    Args:
        folder (str): Folder to check
        start_date (date): Initial date
        end_date (date): Final date
        pattern (str): Datetime format of the folder
    Returns:
        bool: True for success, False otherwise
    """
    try:
        dt = datetime.strptime(folder, pattern)
    except:
        raise ValueError('Folder {} does not match {} format'.format(folder, pattern))
    return start_date <= dt.date() <= end_date



# Copyright (c) 2012-2020 Jicamarca Radio Observatory
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
"""API to create signal chain projects

The API is provide through class: Project
"""

import re
import sys
import ast
import datetime
import traceback
import time
from multiprocessing import Process, Queue
from threading import Thread
from xml.etree.ElementTree import ElementTree, Element, SubElement

from schainpy.admin import Alarm, SchainWarning
from schainpy.model import *
from schainpy.utils import log


class ConfBase():

    def __init__(self):

        self.id = '0'
        self.name = None
        self.priority = None
        self.parameters = {}
        self.object = None
        self.operations = []

    def getId(self):

        return self.id
    
    def getNewId(self):

        return int(self.id) * 10 + len(self.operations) + 1

    def updateId(self, new_id):

        self.id = str(new_id)

        n = 1
        for conf in self.operations:
            conf_id = str(int(new_id) * 10 + n)
            conf.updateId(conf_id)
            n += 1

    def getKwargs(self):

        params = {}

        for key, value in self.parameters.items():
            if value not in (None, '', ' '):
                params[key] = value
        
        return params

    def update(self, **kwargs):

        for key, value in kwargs.items():
            self.addParameter(name=key, value=value)

    def addParameter(self, name, value, format=None):
        '''
        '''

        if isinstance(value, str) and re.search(r'(\d+/\d+/\d+)', value):
            self.parameters[name] = datetime.date(*[int(x) for x in value.split('/')])
        elif isinstance(value, str) and re.search(r'(\d+:\d+:\d+)', value):
            self.parameters[name] = datetime.time(*[int(x) for x in value.split(':')])
        else:
            try:
                self.parameters[name] = ast.literal_eval(value)
            except:
                if isinstance(value, str) and ',' in value:
                    self.parameters[name] = value.split(',')
                else:
                    self.parameters[name] = value

    def getParameters(self):

        params = {}
        for key, value in self.parameters.items():
            s = type(value).__name__
            if s == 'date':
                params[key] = value.strftime('%Y/%m/%d')
            elif s == 'time':
                params[key] = value.strftime('%H:%M:%S')
            else:
                params[key] = str(value)

        return params
    
    def makeXml(self, element):

        xml = SubElement(element, self.ELEMENTNAME)
        for label in self.xml_labels:
            xml.set(label, str(getattr(self, label)))
        
        for key, value in self.getParameters().items():
            xml_param = SubElement(xml, 'Parameter')
            xml_param.set('name', key)
            xml_param.set('value', value)
        
        for conf in self.operations:
            conf.makeXml(xml)
            
    def __str__(self):

        if self.ELEMENTNAME == 'Operation':
            s = '  {}[id={}]\n'.format(self.name, self.id)
        else:
            s = '{}[id={}, inputId={}]\n'.format(self.name, self.id, self.inputId)

        for key, value in self.parameters.items():
            if self.ELEMENTNAME == 'Operation':
                s += '    {}: {}\n'.format(key, value)
            else:
                s += '  {}: {}\n'.format(key, value)
        
        for conf in self.operations:
            s += str(conf)

        return s

class OperationConf(ConfBase):

    ELEMENTNAME = 'Operation'
    xml_labels = ['id', 'name']

    def setup(self, id, name, priority, project_id, err_queue):

        self.id = str(id)
        self.project_id = project_id
        self.name = name
        self.type = 'other'
        self.err_queue = err_queue

    def readXml(self, element, project_id, err_queue):

        self.id = element.get('id')
        self.name = element.get('name')
        self.type = 'other'
        self.project_id = str(project_id)
        self.err_queue = err_queue

        for elm in element.iter('Parameter'):
            self.addParameter(elm.get('name'), elm.get('value'))

    def createObject(self):

        className = eval(self.name)

        if 'Plot' in self.name or 'Writer' in self.name or 'Send' in self.name or 'print' in self.name:
            kwargs = self.getKwargs()
            opObj = className(self.id, self.id, self.project_id, self.err_queue, **kwargs)
            opObj.start()
            self.type = 'external'
        else:
            opObj = className()

        self.object = opObj
        return opObj

class ProcUnitConf(ConfBase):

    ELEMENTNAME = 'ProcUnit'
    xml_labels = ['id', 'inputId', 'name']

    def setup(self, project_id, id, name, datatype, inputId, err_queue):
        '''
        '''
        
        if datatype == None and name == None:
            raise ValueError('datatype or name should be defined')

        if name == None:
            if 'Proc' in datatype:
                name = datatype
            else:
                name = '%sProc' % (datatype)

        if datatype == None:
            datatype = name.replace('Proc', '')

        self.id = str(id)
        self.project_id = project_id
        self.name = name
        self.datatype = datatype
        self.inputId = inputId
        self.err_queue = err_queue
        self.operations = []
        self.parameters = {}

    def removeOperation(self, id):

        i = [1 if x.id==id else 0 for x in self.operations]
        self.operations.pop(i.index(1))
        
    def getOperation(self, id):

        for conf in self.operations:
            if conf.id == id:
                return conf

    def addOperation(self, name, optype='self'):
        '''
        '''

        id = self.getNewId()
        conf = OperationConf()
        conf.setup(id, name=name, priority='0', project_id=self.project_id, err_queue=self.err_queue)
        self.operations.append(conf)

        return conf

    def readXml(self, element, project_id, err_queue):

        self.id = element.get('id')
        self.name = element.get('name')
        self.inputId = None if element.get('inputId') == 'None' else element.get('inputId')
        self.datatype = element.get('datatype', self.name.replace(self.ELEMENTNAME.replace('Unit', ''), ''))
        self.project_id = str(project_id)
        self.err_queue = err_queue
        self.operations = []
        self.parameters = {}
        
        for elm in element:
            if elm.tag == 'Parameter':
                self.addParameter(elm.get('name'), elm.get('value'))
            elif elm.tag == 'Operation':
                conf = OperationConf()
                conf.readXml(elm, project_id, err_queue)
                self.operations.append(conf)

    def createObjects(self):
        '''
        Instancia de unidades de procesamiento.
        '''

        className = eval(self.name)
        kwargs = self.getKwargs()
        procUnitObj = className()
        procUnitObj.name = self.name
        log.success('creating process...', self.name)

        for conf in self.operations:
            
            opObj = conf.createObject()
            
            log.success('adding operation: {}, type:{}'.format(
                conf.name,
                conf.type), self.name)
            
            procUnitObj.addOperation(conf, opObj)
     
        self.object = procUnitObj

    def run(self):
        '''
        '''
        
        return self.object.call(**self.getKwargs())


class ReadUnitConf(ProcUnitConf):

    ELEMENTNAME = 'ReadUnit'

    def __init__(self):

        self.id = None
        self.datatype = None
        self.name = None
        self.inputId = None
        self.operations = []
        self.parameters = {}
    
    def setup(self, project_id, id, name, datatype, err_queue, path='', startDate='', endDate='',
              startTime='', endTime='', server=None, **kwargs):
        
        if datatype == None and name == None:
            raise ValueError('datatype or name should be defined')
        if name == None:
            if 'Reader' in datatype:
                name = datatype
                datatype = name.replace('Reader','')
            else:
                name = '{}Reader'.format(datatype)
        if datatype == None:
            if 'Reader' in name:
                datatype = name.replace('Reader','')
            else:
                datatype = name
                name = '{}Reader'.format(name)

        self.id = id
        self.project_id = project_id
        self.name = name
        self.datatype = datatype
        self.err_queue = err_queue        
        
        self.addParameter(name='path', value=path)
        self.addParameter(name='startDate', value=startDate)
        self.addParameter(name='endDate', value=endDate)
        self.addParameter(name='startTime', value=startTime)
        self.addParameter(name='endTime', value=endTime)

        for key, value in kwargs.items():
            self.addParameter(name=key, value=value)


class Project(Process):
    """API to create signal chain projects"""

    ELEMENTNAME = 'Project'

    def __init__(self, name=''):

        Process.__init__(self)
        self.id = '1'
        if name:
            self.name = '{} ({})'.format(Process.__name__, name)
        self.filename = None
        self.description = None
        self.email = None
        self.alarm = []
        self.configurations = {}
        # self.err_queue = Queue()
        self.err_queue = None
        self.started = False

    def getNewId(self):

        idList = list(self.configurations.keys())
        id = int(self.id) * 10

        while True:
            id += 1

            if str(id) in idList:
                continue

            break

        return str(id)

    def updateId(self, new_id):

        self.id = str(new_id)

        keyList = list(self.configurations.keys())
        keyList.sort()

        n = 1
        new_confs = {}

        for procKey in keyList:

            conf = self.configurations[procKey]
            idProcUnit = str(int(self.id) * 10 + n)
            conf.updateId(idProcUnit)
            new_confs[idProcUnit] = conf
            n += 1

        self.configurations = new_confs

    def setup(self, id=1, name='', description='', email=None, alarm=[]):

        self.id = str(id)
        self.description = description 
        self.email = email
        self.alarm = alarm
        if name:
            self.name = '{} ({})'.format(Process.__name__, name)

    def update(self, **kwargs):

        for key, value in kwargs.items():
            setattr(self, key, value)

    def clone(self):

        p = Project()
        p.id = self.id
        p.name = self.name
        p.description = self.description
        p.configurations = self.configurations.copy()

        return p

    def addReadUnit(self, id=None, datatype=None, name=None, **kwargs):

        '''
        '''

        if id is None:
            idReadUnit = self.getNewId()
        else:
            idReadUnit = str(id)

        conf = ReadUnitConf()
        conf.setup(self.id, idReadUnit, name, datatype, self.err_queue, **kwargs)
        self.configurations[conf.id] = conf
        
        return conf

    def addProcUnit(self, id=None, inputId='0', datatype=None, name=None):

        '''
        '''

        if id is None:
            idProcUnit = self.getNewId()
        else:
            idProcUnit = id
        
        conf = ProcUnitConf()
        conf.setup(self.id, idProcUnit, name, datatype, inputId, self.err_queue)
        self.configurations[conf.id] = conf

        return conf

    def removeProcUnit(self, id):

        if id in self.configurations:
            self.configurations.pop(id)

    def getReadUnit(self):

        for obj in list(self.configurations.values()):
            if obj.ELEMENTNAME == 'ReadUnit':
                return obj

        return None

    def getProcUnit(self, id):

        return self.configurations[id]

    def getUnits(self):

        keys = list(self.configurations)
        keys.sort()

        for key in keys:
            yield self.configurations[key]

    def updateUnit(self, id, **kwargs):

        conf = self.configurations[id].update(**kwargs)
    
    def makeXml(self):

        xml = Element('Project')
        xml.set('id', str(self.id))
        xml.set('name', self.name)
        xml.set('description', self.description)

        for conf in self.configurations.values():
            conf.makeXml(xml)

        self.xml = xml

    def writeXml(self, filename=None):

        if filename == None:
            if self.filename:
                filename = self.filename
            else:
                filename = 'schain.xml'

        if not filename:
            print('filename has not been defined. Use setFilename(filename) for do it.')
            return 0

        abs_file = os.path.abspath(filename)

        if not os.access(os.path.dirname(abs_file), os.W_OK):
            print('No write permission on %s' % os.path.dirname(abs_file))
            return 0

        if os.path.isfile(abs_file) and not(os.access(abs_file, os.W_OK)):
            print('File %s already exists and it could not be overwriten' % abs_file)
            return 0

        self.makeXml()

        ElementTree(self.xml).write(abs_file, method='xml')

        self.filename = abs_file

        return 1

    def readXml(self, filename):

        abs_file = os.path.abspath(filename)

        self.configurations = {}

        try:
            self.xml = ElementTree().parse(abs_file)
        except:
            log.error('Error reading %s, verify file format' % filename)
            return 0

        self.id = self.xml.get('id')
        self.name = self.xml.get('name')
        self.description = self.xml.get('description')

        for element in self.xml:
            if element.tag == 'ReadUnit':
                conf = ReadUnitConf()
                conf.readXml(element, self.id, self.err_queue)
                self.configurations[conf.id] = conf
            elif element.tag == 'ProcUnit':
                conf = ProcUnitConf()
                input_proc = self.configurations[element.get('inputId')]
                conf.readXml(element, self.id, self.err_queue)
                self.configurations[conf.id] = conf

        self.filename = abs_file
        
        return 1

    def __str__(self):

        text = '\nProject[id=%s, name=%s, description=%s]\n\n' % (
            self.id,
            self.name,
            self.description,
            )

        for conf in self.configurations.values():
            text += '{}'.format(conf)

        return text

    def createObjects(self):

        keys = list(self.configurations.keys())
        keys.sort()
        for key in keys:
            conf = self.configurations[key]
            conf.createObjects()
            if conf.inputId is not None:
                conf.object.setInput(self.configurations[conf.inputId].object)

    def monitor(self):

        t = Thread(target=self._monitor, args=(self.err_queue, self.ctx))
        t.start()
    
    def _monitor(self, queue, ctx):

        import socket
        
        procs = 0
        err_msg = ''
        
        while True:
            msg = queue.get()
            if '#_start_#' in msg:
                procs += 1
            elif '#_end_#' in msg:
                procs -=1
            else:
                err_msg = msg
            
            if procs == 0 or 'Traceback' in err_msg:                
                break
            time.sleep(0.1)
        
        if '|' in err_msg:
            name, err = err_msg.split('|')
            if 'SchainWarning' in err:
                log.warning(err.split('SchainWarning:')[-1].split('\n')[0].strip(), name)
            elif 'SchainError' in err:
                log.error(err.split('SchainError:')[-1].split('\n')[0].strip(), name)
            else:
                log.error(err, name)
        else:            
            name, err = self.name, err_msg
        
        time.sleep(1)
            
        ctx.term()

        message = ''.join(err)

        if err_msg:
            subject = 'SChain v%s: Error running %s\n' % (
                schainpy.__version__, self.name)

            subtitle = 'Hostname: %s\n' % socket.gethostbyname(
                socket.gethostname())
            subtitle += 'Working directory: %s\n' % os.path.abspath('./')
            subtitle += 'Configuration file: %s\n' % self.filename
            subtitle += 'Time: %s\n' % str(datetime.datetime.now())

            readUnitConfObj = self.getReadUnit()
            if readUnitConfObj:
                subtitle += '\nInput parameters:\n'
                subtitle += '[Data path = %s]\n' % readUnitConfObj.parameters['path']
                subtitle += '[Start date = %s]\n' % readUnitConfObj.parameters['startDate']
                subtitle += '[End date = %s]\n' % readUnitConfObj.parameters['endDate']
                subtitle += '[Start time = %s]\n' % readUnitConfObj.parameters['startTime']
                subtitle += '[End time = %s]\n' % readUnitConfObj.parameters['endTime']

            a = Alarm(
                modes=self.alarm, 
                email=self.email,
                message=message,
                subject=subject,
                subtitle=subtitle,
                filename=self.filename
            )

            a.start()

    def setFilename(self, filename):

        self.filename = filename

    def runProcs(self):

        err = False
        n = len(self.configurations)
        
        while not err:
            for conf in self.getUnits():
                ok = conf.run()                
                if ok is 'Error':
                    n -= 1
                    continue
                elif not ok:
                    break
            if n == 0:
                err = True
        
    def run(self):

        log.success('\nStarting Project {} [id={}]'.format(self.name, self.id), tag='')
        self.started = True
        self.start_time = time.time()        
        self.createObjects()
        self.runProcs()
        log.success('{} Done (Time: {:4.2f}s)'.format(
            self.name,
            time.time()-self.start_time), '')



'''

$Author: murco $
$Id: JRODataIO.py 169 2012-11-19 21:57:03Z murco $
'''

from .jroIO_voltage import *
from .jroIO_spectra import *
from .jroIO_heispectra import *
from .jroIO_usrp import *
from .jroIO_digitalRF import *
from .jroIO_kamisr import *
from .jroIO_param import *
from .jroIO_hf import *

from .jroIO_madrigal import *

from .bltrIO_param import *
from .bltrIO_spectra import *
from .jroIO_mira35c import *
from .julIO_param import *

from .pxIO_param import *
from .jroIO_simulator import *


import time
import itertools

import numpy

from schainpy.model.proc.jroproc_base import ProcessingUnit, MPDecorator, Operation
from schainpy.model.data.jrodata import Spectra
from schainpy.model.data.jrodata import hildebrand_sekhon
from schainpy.utils import log


class SpectraProc(ProcessingUnit):


    def __init__(self):

        ProcessingUnit.__init__(self)

        self.buffer = None
        self.firstdatatime = None
        self.profIndex = 0
        self.dataOut = Spectra()
        self.id_min = None
        self.id_max = None
        self.setupReq = False #Agregar a todas las unidades de proc

    def __updateSpecFromVoltage(self):

        self.dataOut.timeZone = self.dataIn.timeZone
        self.dataOut.dstFlag = self.dataIn.dstFlag
        self.dataOut.errorCount = self.dataIn.errorCount
        self.dataOut.useLocalTime = self.dataIn.useLocalTime
        try:
            self.dataOut.processingHeaderObj = self.dataIn.processingHeaderObj.copy()
        except:
            pass
        self.dataOut.radarControllerHeaderObj = self.dataIn.radarControllerHeaderObj.copy()
        self.dataOut.systemHeaderObj = self.dataIn.systemHeaderObj.copy()
        self.dataOut.channelList = self.dataIn.channelList
        self.dataOut.heightList = self.dataIn.heightList
        self.dataOut.dtype = numpy.dtype([('real', '<f4'), ('imag', '<f4')])

        self.dataOut.nBaud = self.dataIn.nBaud
        self.dataOut.nCode = self.dataIn.nCode
        self.dataOut.code = self.dataIn.code
        self.dataOut.nProfiles = self.dataOut.nFFTPoints

        self.dataOut.flagDiscontinuousBlock = self.dataIn.flagDiscontinuousBlock
        self.dataOut.utctime = self.firstdatatime
        # asumo q la data esta decodificada
        self.dataOut.flagDecodeData = self.dataIn.flagDecodeData
        # asumo q la data esta sin flip
        self.dataOut.flagDeflipData = self.dataIn.flagDeflipData
        self.dataOut.flagShiftFFT = False

        self.dataOut.nCohInt = self.dataIn.nCohInt
        self.dataOut.nIncohInt = 1

        self.dataOut.windowOfFilter = self.dataIn.windowOfFilter

        self.dataOut.frequency = self.dataIn.frequency
        self.dataOut.realtime = self.dataIn.realtime

        self.dataOut.azimuth = self.dataIn.azimuth
        self.dataOut.zenith = self.dataIn.zenith

        self.dataOut.beam.codeList = self.dataIn.beam.codeList
        self.dataOut.beam.azimuthList = self.dataIn.beam.azimuthList
        self.dataOut.beam.zenithList = self.dataIn.beam.zenithList

    def __getFft(self):
        """
        Convierte valores de Voltaje a Spectra

        Affected:
            self.dataOut.data_spc
            self.dataOut.data_cspc
            self.dataOut.data_dc
            self.dataOut.heightList
            self.profIndex
            self.buffer
            self.dataOut.flagNoData
        """
        fft_volt = numpy.fft.fft(
            self.buffer, n=self.dataOut.nFFTPoints, axis=1)
        fft_volt = fft_volt.astype(numpy.dtype('complex'))
        dc = fft_volt[:, 0, :]

        # calculo de self-spectra
        fft_volt = numpy.fft.fftshift(fft_volt, axes=(1,))
        spc = fft_volt * numpy.conjugate(fft_volt)
        spc = spc.real

        blocksize = 0
        blocksize += dc.size
        blocksize += spc.size

        cspc = None
        pairIndex = 0
        if self.dataOut.pairsList != None:
            # calculo de cross-spectra
            cspc = numpy.zeros(
                (self.dataOut.nPairs, self.dataOut.nFFTPoints, self.dataOut.nHeights), dtype='complex')
            for pair in self.dataOut.pairsList:
                if pair[0] not in self.dataOut.channelList:
                    raise ValueError("Error getting CrossSpectra: pair 0 of %s is not in channelList = %s" % (
                        str(pair), str(self.dataOut.channelList)))
                if pair[1] not in self.dataOut.channelList:
                    raise ValueError("Error getting CrossSpectra: pair 1 of %s is not in channelList = %s" % (
                        str(pair), str(self.dataOut.channelList)))

                cspc[pairIndex, :, :] = fft_volt[pair[0], :, :] * \
                    numpy.conjugate(fft_volt[pair[1], :, :])
                pairIndex += 1
            blocksize += cspc.size

        self.dataOut.data_spc = spc
        self.dataOut.data_cspc = cspc
        self.dataOut.data_dc = dc
        self.dataOut.blockSize = blocksize
        self.dataOut.flagShiftFFT = False

    def run(self, nProfiles=None, nFFTPoints=None, pairsList=[], ippFactor=None, shift_fft=False):
        
        if self.dataIn.type == "Spectra":
            self.dataOut.copy(self.dataIn)
            if shift_fft:
                #desplaza a la derecha en el eje 2 determinadas posiciones
                shift = int(self.dataOut.nFFTPoints/2)
                self.dataOut.data_spc = numpy.roll(self.dataOut.data_spc, shift , axis=1)

                if self.dataOut.data_cspc is not None:
                    #desplaza a la derecha en el eje 2 determinadas posiciones
                    self.dataOut.data_cspc = numpy.roll(self.dataOut.data_cspc, shift, axis=1)

            return True

        if self.dataIn.type == "Voltage":

            self.dataOut.flagNoData = True

            if nFFTPoints == None:
                raise ValueError("This SpectraProc.run() need nFFTPoints input variable")

            if nProfiles == None:
                nProfiles = nFFTPoints

            if ippFactor == None:
                ippFactor = 1

            self.dataOut.ippFactor = ippFactor

            self.dataOut.nFFTPoints = nFFTPoints
            self.dataOut.pairsList = pairsList

            if self.buffer is None:
                self.buffer = numpy.zeros((self.dataIn.nChannels,
                                           nProfiles,
                                           self.dataIn.nHeights),
                                          dtype='complex')

            if self.dataIn.flagDataAsBlock:
                nVoltProfiles = self.dataIn.data.shape[1]

                if nVoltProfiles == nProfiles:
                    self.buffer = self.dataIn.data.copy()
                    self.profIndex = nVoltProfiles

                elif nVoltProfiles < nProfiles:

                    if self.profIndex == 0:
                        self.id_min = 0
                        self.id_max = nVoltProfiles

                    self.buffer[:, self.id_min:self.id_max,
                                :] = self.dataIn.data
                    self.profIndex += nVoltProfiles
                    self.id_min += nVoltProfiles
                    self.id_max += nVoltProfiles
                else:
                    raise ValueError("The type object %s has %d profiles, it should just has %d profiles" % (
                        self.dataIn.type, self.dataIn.data.shape[1], nProfiles))
                    self.dataOut.flagNoData = True
                    return 0
            else:
                self.buffer[:, self.profIndex, :] = self.dataIn.data.copy()
                self.profIndex += 1

            if self.firstdatatime == None:
                self.firstdatatime = self.dataIn.utctime

            if self.profIndex == nProfiles:
                self.__updateSpecFromVoltage()
                self.__getFft()

                self.dataOut.flagNoData = False
                self.firstdatatime = None
                self.profIndex = 0

            return True

        raise ValueError("The type of input object '%s' is not valid" % (
            self.dataIn.type))

    def __selectPairs(self, pairsList):

        if not pairsList:
            return

        pairs = []
        pairsIndex = []

        for pair in pairsList:
            if pair[0] not in self.dataOut.channelList or pair[1] not in self.dataOut.channelList:
                continue
            pairs.append(pair)
            pairsIndex.append(pairs.index(pair))

        self.dataOut.data_cspc = self.dataOut.data_cspc[pairsIndex]
        self.dataOut.pairsList = pairs

        return
    
    def selectFFTs(self, minFFT, maxFFT ):
        """
        Selecciona un bloque de datos en base a un grupo de valores de puntos FFTs segun el rango 
        minFFT<= FFT <= maxFFT
        """
        
        if (minFFT > maxFFT):
            raise ValueError("Error selecting heights: Height range (%d,%d) is not valid" % (minFFT, maxFFT))

        if (minFFT < self.dataOut.getFreqRange()[0]):
            minFFT = self.dataOut.getFreqRange()[0]

        if (maxFFT > self.dataOut.getFreqRange()[-1]):
            maxFFT = self.dataOut.getFreqRange()[-1]

        minIndex = 0
        maxIndex = 0
        FFTs = self.dataOut.getFreqRange()

        inda = numpy.where(FFTs >= minFFT)
        indb = numpy.where(FFTs <= maxFFT)

        try:
            minIndex = inda[0][0]
        except:
            minIndex = 0

        try:
            maxIndex = indb[0][-1]
        except:
            maxIndex = len(FFTs)

        self.selectFFTsByIndex(minIndex, maxIndex)

        return 1
    
    def getBeaconSignal(self, tauindex=0, channelindex=0, hei_ref=None):
        newheis = numpy.where(
            self.dataOut.heightList > self.dataOut.radarControllerHeaderObj.Taus[tauindex])

        if hei_ref != None:
            newheis = numpy.where(self.dataOut.heightList > hei_ref)

        minIndex = min(newheis[0])
        maxIndex = max(newheis[0])
        data_spc = self.dataOut.data_spc[:, :, minIndex:maxIndex + 1]
        heightList = self.dataOut.heightList[minIndex:maxIndex + 1]

        # determina indices
        nheis = int(self.dataOut.radarControllerHeaderObj.txB /
                    (self.dataOut.heightList[1] - self.dataOut.heightList[0]))
        avg_dB = 10 * \
            numpy.log10(numpy.sum(data_spc[channelindex, :, :], axis=0))
        beacon_dB = numpy.sort(avg_dB)[-nheis:]
        beacon_heiIndexList = []
        for val in avg_dB.tolist():
            if val >= beacon_dB[0]:
                beacon_heiIndexList.append(avg_dB.tolist().index(val))

        #data_spc = data_spc[:,:,beacon_heiIndexList]
        data_cspc = None
        if self.dataOut.data_cspc is not None:
            data_cspc = self.dataOut.data_cspc[:, :, minIndex:maxIndex + 1]
            #data_cspc = data_cspc[:,:,beacon_heiIndexList]

        data_dc = None
        if self.dataOut.data_dc is not None:
            data_dc = self.dataOut.data_dc[:, minIndex:maxIndex + 1]
            #data_dc = data_dc[:,beacon_heiIndexList]

        self.dataOut.data_spc = data_spc
        self.dataOut.data_cspc = data_cspc
        self.dataOut.data_dc = data_dc
        self.dataOut.heightList = heightList
        self.dataOut.beacon_heiIndexList = beacon_heiIndexList

        return 1

    def selectFFTsByIndex(self, minIndex, maxIndex):
        """
        
        """

        if (minIndex < 0) or (minIndex > maxIndex):
            raise ValueError("Error selecting heights: Index range (%d,%d) is not valid" % (minIndex, maxIndex))

        if (maxIndex >= self.dataOut.nProfiles):
            maxIndex = self.dataOut.nProfiles-1

        #Spectra
        data_spc = self.dataOut.data_spc[:,minIndex:maxIndex+1,:]

        data_cspc = None
        if self.dataOut.data_cspc is not None:
            data_cspc = self.dataOut.data_cspc[:,minIndex:maxIndex+1,:]

        data_dc = None
        if self.dataOut.data_dc is not None:
            data_dc = self.dataOut.data_dc[minIndex:maxIndex+1,:]

        self.dataOut.data_spc = data_spc
        self.dataOut.data_cspc = data_cspc
        self.dataOut.data_dc = data_dc
        
        self.dataOut.ippSeconds = self.dataOut.ippSeconds*(self.dataOut.nFFTPoints / numpy.shape(data_cspc)[1])
        self.dataOut.nFFTPoints = numpy.shape(data_cspc)[1]
        self.dataOut.profilesPerBlock = numpy.shape(data_cspc)[1]

        return 1

    def getNoise(self, minHei=None, maxHei=None, minVel=None, maxVel=None):
        # validacion de rango
        if minHei == None:
            minHei = self.dataOut.heightList[0]

        if maxHei == None:
            maxHei = self.dataOut.heightList[-1]

        if (minHei < self.dataOut.heightList[0]) or (minHei > maxHei):
            print('minHei: %.2f is out of the heights range' % (minHei))
            print('minHei is setting to %.2f' % (self.dataOut.heightList[0]))
            minHei = self.dataOut.heightList[0]

        if (maxHei > self.dataOut.heightList[-1]) or (maxHei < minHei):
            print('maxHei: %.2f is out of the heights range' % (maxHei))
            print('maxHei is setting to %.2f' % (self.dataOut.heightList[-1]))
            maxHei = self.dataOut.heightList[-1]

        # validacion de velocidades
        velrange = self.dataOut.getVelRange(1)

        if minVel == None:
            minVel = velrange[0]

        if maxVel == None:
            maxVel = velrange[-1]

        if (minVel < velrange[0]) or (minVel > maxVel):
            print('minVel: %.2f is out of the velocity range' % (minVel))
            print('minVel is setting to %.2f' % (velrange[0]))
            minVel = velrange[0]

        if (maxVel > velrange[-1]) or (maxVel < minVel):
            print('maxVel: %.2f is out of the velocity range' % (maxVel))
            print('maxVel is setting to %.2f' % (velrange[-1]))
            maxVel = velrange[-1]

        # seleccion de indices para rango
        minIndex = 0
        maxIndex = 0
        heights = self.dataOut.heightList

        inda = numpy.where(heights >= minHei)
        indb = numpy.where(heights <= maxHei)

        try:
            minIndex = inda[0][0]
        except:
            minIndex = 0

        try:
            maxIndex = indb[0][-1]
        except:
            maxIndex = len(heights)

        if (minIndex < 0) or (minIndex > maxIndex):
            raise ValueError("some value in (%d,%d) is not valid" % (
                minIndex, maxIndex))

        if (maxIndex >= self.dataOut.nHeights):
            maxIndex = self.dataOut.nHeights - 1

        # seleccion de indices para velocidades
        indminvel = numpy.where(velrange >= minVel)
        indmaxvel = numpy.where(velrange <= maxVel)
        try:
            minIndexVel = indminvel[0][0]
        except:
            minIndexVel = 0

        try:
            maxIndexVel = indmaxvel[0][-1]
        except:
            maxIndexVel = len(velrange)

        # seleccion del espectro
        data_spc = self.dataOut.data_spc[:,
                                         minIndexVel:maxIndexVel + 1, minIndex:maxIndex + 1]
        # estimacion de ruido
        noise = numpy.zeros(self.dataOut.nChannels)

        for channel in range(self.dataOut.nChannels):
            daux = data_spc[channel, :, :]
            sortdata = numpy.sort(daux, axis=None)
            noise[channel] = hildebrand_sekhon(sortdata, self.dataOut.nIncohInt)

        self.dataOut.noise_estimation = noise.copy()

        return 1

class removeDC(Operation):

    def run(self, dataOut, mode=2):
        self.dataOut = dataOut
        jspectra = self.dataOut.data_spc
        jcspectra = self.dataOut.data_cspc

        num_chan = jspectra.shape[0]
        num_hei = jspectra.shape[2]

        if jcspectra is not None:
            jcspectraExist = True
            num_pairs = jcspectra.shape[0]
        else:
            jcspectraExist = False

        freq_dc = int(jspectra.shape[1] / 2)
        ind_vel = numpy.array([-2, -1, 1, 2]) + freq_dc
        ind_vel = ind_vel.astype(int)

        if ind_vel[0] < 0:
            ind_vel[list(range(0, 1))] = ind_vel[list(range(0, 1))] + self.num_prof

        if mode == 1:
            jspectra[:, freq_dc, :] = (
                jspectra[:, ind_vel[1], :] + jspectra[:, ind_vel[2], :]) / 2  # CORRECCION

            if jcspectraExist:
                jcspectra[:, freq_dc, :] = (
                    jcspectra[:, ind_vel[1], :] + jcspectra[:, ind_vel[2], :]) / 2

        if mode == 2:

            vel = numpy.array([-2, -1, 1, 2])
            xx = numpy.zeros([4, 4])

            for fil in range(4):
                xx[fil, :] = vel[fil]**numpy.asarray(list(range(4)))

            xx_inv = numpy.linalg.inv(xx)
            xx_aux = xx_inv[0, :]

            for ich in range(num_chan):                
                yy = jspectra[ich, ind_vel, :]
                jspectra[ich, freq_dc, :] = numpy.dot(xx_aux, yy)

                junkid = jspectra[ich, freq_dc, :] <= 0
                cjunkid = sum(junkid)

                if cjunkid.any():
                    jspectra[ich, freq_dc, junkid.nonzero()] = (
                        jspectra[ich, ind_vel[1], junkid] + jspectra[ich, ind_vel[2], junkid]) / 2

            if jcspectraExist:
                for ip in range(num_pairs):
                    yy = jcspectra[ip, ind_vel, :]
                    jcspectra[ip, freq_dc, :] = numpy.dot(xx_aux, yy)

        self.dataOut.data_spc = jspectra
        self.dataOut.data_cspc = jcspectra

        return self.dataOut

class removeInterference(Operation):

    def removeInterference2(self):
        
        cspc = self.dataOut.data_cspc
        spc = self.dataOut.data_spc
        Heights = numpy.arange(cspc.shape[2]) 
        realCspc = numpy.abs(cspc)
        
        for i in range(cspc.shape[0]):
            LinePower= numpy.sum(realCspc[i], axis=0)
            Threshold = numpy.amax(LinePower)-numpy.sort(LinePower)[len(Heights)-int(len(Heights)*0.1)]
            SelectedHeights = Heights[ numpy.where( LinePower < Threshold ) ]
            InterferenceSum = numpy.sum( realCspc[i,:,SelectedHeights], axis=0 )
            InterferenceThresholdMin = numpy.sort(InterferenceSum)[int(len(InterferenceSum)*0.98)]
            InterferenceThresholdMax = numpy.sort(InterferenceSum)[int(len(InterferenceSum)*0.99)]
            
            
            InterferenceRange = numpy.where( ([InterferenceSum > InterferenceThresholdMin]))# , InterferenceSum < InterferenceThresholdMax]) )
            #InterferenceRange = numpy.where( ([InterferenceRange < InterferenceThresholdMax]))
            if len(InterferenceRange)<int(cspc.shape[1]*0.3):
                cspc[i,InterferenceRange,:] = numpy.NaN
            
        self.dataOut.data_cspc = cspc
        
    def removeInterference(self, interf = 2, hei_interf = None, nhei_interf = None, offhei_interf = None):

        jspectra = self.dataOut.data_spc
        jcspectra = self.dataOut.data_cspc
        jnoise = self.dataOut.getNoise()
        num_incoh = self.dataOut.nIncohInt

        num_channel = jspectra.shape[0]
        num_prof = jspectra.shape[1]
        num_hei = jspectra.shape[2]

        # hei_interf
        if hei_interf is None:
            count_hei = int(num_hei / 2)
            hei_interf = numpy.asmatrix(list(range(count_hei))) + num_hei - count_hei
            hei_interf = numpy.asarray(hei_interf)[0]
        # nhei_interf
        if (nhei_interf == None):
            nhei_interf = 5
        if (nhei_interf < 1):
            nhei_interf = 1
        if (nhei_interf > count_hei):
            nhei_interf = count_hei
        if (offhei_interf == None):
            offhei_interf = 0

        ind_hei = list(range(num_hei))
#         mask_prof = numpy.asarray(range(num_prof - 2)) + 1
#         mask_prof[range(num_prof/2 - 1,len(mask_prof))] += 1
        mask_prof = numpy.asarray(list(range(num_prof)))
        num_mask_prof = mask_prof.size
        comp_mask_prof = [0, num_prof / 2]

        # noise_exist:    Determina si la variable jnoise ha sido definida y contiene la informacion del ruido de cada canal
        if (jnoise.size < num_channel or numpy.isnan(jnoise).any()):
            jnoise = numpy.nan
        noise_exist = jnoise[0] < numpy.Inf

        # Subrutina de Remocion de la Interferencia
        for ich in range(num_channel):
            # Se ordena los espectros segun su potencia (menor a mayor)
            power = jspectra[ich, mask_prof, :]
            power = power[:, hei_interf]
            power = power.sum(axis=0)
            psort = power.ravel().argsort()

            # Se estima la interferencia promedio en los Espectros de Potencia empleando
            junkspc_interf = jspectra[ich, :, hei_interf[psort[list(range(
                offhei_interf, nhei_interf + offhei_interf))]]]

            if noise_exist:
                #    tmp_noise = jnoise[ich] / num_prof
                tmp_noise = jnoise[ich]
            junkspc_interf = junkspc_interf - tmp_noise
            #junkspc_interf[:,comp_mask_prof] = 0

            jspc_interf = junkspc_interf.sum(axis=0) / nhei_interf
            jspc_interf = jspc_interf.transpose()
            # Calculando el espectro de interferencia promedio
            noiseid = numpy.where(
                jspc_interf <= tmp_noise / numpy.sqrt(num_incoh))
            noiseid = noiseid[0]
            cnoiseid = noiseid.size
            interfid = numpy.where(
                jspc_interf > tmp_noise / numpy.sqrt(num_incoh))
            interfid = interfid[0]
            cinterfid = interfid.size

            if (cnoiseid > 0):
                jspc_interf[noiseid] = 0

            # Expandiendo los perfiles a limpiar
            if (cinterfid > 0):
                new_interfid = (
                    numpy.r_[interfid - 1, interfid, interfid + 1] + num_prof) % num_prof
                new_interfid = numpy.asarray(new_interfid)
                new_interfid = {x for x in new_interfid}
                new_interfid = numpy.array(list(new_interfid))
                new_cinterfid = new_interfid.size
            else:
                new_cinterfid = 0

            for ip in range(new_cinterfid):
                ind = junkspc_interf[:, new_interfid[ip]].ravel().argsort()
                jspc_interf[new_interfid[ip]
                            ] = junkspc_interf[ind[nhei_interf // 2], new_interfid[ip]]

            jspectra[ich, :, ind_hei] = jspectra[ich, :,
                                                 ind_hei] - jspc_interf  # Corregir indices

            # Removiendo la interferencia del punto de mayor interferencia
            ListAux = jspc_interf[mask_prof].tolist()
            maxid = ListAux.index(max(ListAux))

            if cinterfid > 0:
                for ip in range(cinterfid * (interf == 2) - 1):
                    ind = (jspectra[ich, interfid[ip], :] < tmp_noise *
                           (1 + 1 / numpy.sqrt(num_incoh))).nonzero()
                    cind = len(ind)

                    if (cind > 0):
                        jspectra[ich, interfid[ip], ind] = tmp_noise * \
                            (1 + (numpy.random.uniform(cind) - 0.5) /
                             numpy.sqrt(num_incoh))

                ind = numpy.array([-2, -1, 1, 2])
                xx = numpy.zeros([4, 4])

                for id1 in range(4):
                    xx[:, id1] = ind[id1]**numpy.asarray(list(range(4)))

                xx_inv = numpy.linalg.inv(xx)
                xx = xx_inv[:, 0]
                ind = (ind + maxid + num_mask_prof) % num_mask_prof
                yy = jspectra[ich, mask_prof[ind], :]
                jspectra[ich, mask_prof[maxid], :] = numpy.dot(
                    yy.transpose(), xx)

            indAux = (jspectra[ich, :, :] < tmp_noise *
                      (1 - 1 / numpy.sqrt(num_incoh))).nonzero()
            jspectra[ich, indAux[0], indAux[1]] = tmp_noise * \
                (1 - 1 / numpy.sqrt(num_incoh))

        # Remocion de Interferencia en el Cross Spectra
        if jcspectra is None:
            return jspectra, jcspectra
        num_pairs = int(jcspectra.size / (num_prof * num_hei))
        jcspectra = jcspectra.reshape(num_pairs, num_prof, num_hei)

        for ip in range(num_pairs):

            #-------------------------------------------

            cspower = numpy.abs(jcspectra[ip, mask_prof, :])
            cspower = cspower[:, hei_interf]
            cspower = cspower.sum(axis=0)

            cspsort = cspower.ravel().argsort()
            junkcspc_interf = jcspectra[ip, :, hei_interf[cspsort[list(range(
                offhei_interf, nhei_interf + offhei_interf))]]]
            junkcspc_interf = junkcspc_interf.transpose()
            jcspc_interf = junkcspc_interf.sum(axis=1) / nhei_interf

            ind = numpy.abs(jcspc_interf[mask_prof]).ravel().argsort()

            median_real = int(numpy.median(numpy.real(
                junkcspc_interf[mask_prof[ind[list(range(3 * num_prof // 4))]], :])))
            median_imag = int(numpy.median(numpy.imag(
                junkcspc_interf[mask_prof[ind[list(range(3 * num_prof // 4))]], :])))
            comp_mask_prof = [int(e) for e in comp_mask_prof]
            junkcspc_interf[comp_mask_prof, :] = numpy.complex(
                median_real, median_imag)

            for iprof in range(num_prof):
                ind = numpy.abs(junkcspc_interf[iprof, :]).ravel().argsort()
                jcspc_interf[iprof] = junkcspc_interf[iprof, ind[nhei_interf // 2]]

            # Removiendo la Interferencia
            jcspectra[ip, :, ind_hei] = jcspectra[ip,
                                                  :, ind_hei] - jcspc_interf

            ListAux = numpy.abs(jcspc_interf[mask_prof]).tolist()
            maxid = ListAux.index(max(ListAux))

            ind = numpy.array([-2, -1, 1, 2])
            xx = numpy.zeros([4, 4])

            for id1 in range(4):
                xx[:, id1] = ind[id1]**numpy.asarray(list(range(4)))

            xx_inv = numpy.linalg.inv(xx)
            xx = xx_inv[:, 0]

            ind = (ind + maxid + num_mask_prof) % num_mask_prof
            yy = jcspectra[ip, mask_prof[ind], :]
            jcspectra[ip, mask_prof[maxid], :] = numpy.dot(yy.transpose(), xx)

        # Guardar Resultados
        self.dataOut.data_spc = jspectra
        self.dataOut.data_cspc = jcspectra

        return 1

    def run(self, dataOut, interf = 2,hei_interf = None, nhei_interf = None, offhei_interf = None, mode=1):

        self.dataOut = dataOut

        if mode == 1:
            self.removeInterference(interf = 2,hei_interf = None, nhei_interf = None, offhei_interf = None)
        elif mode == 2:
            self.removeInterference2()

        return self.dataOut


class IncohInt(Operation):

    __profIndex = 0
    __withOverapping = False

    __byTime = False
    __initime = None
    __lastdatatime = None
    __integrationtime = None

    __buffer_spc = None
    __buffer_cspc = None
    __buffer_dc = None

    __dataReady = False

    __timeInterval = None

    n = None

    def __init__(self):

        Operation.__init__(self)

    def setup(self, n=None, timeInterval=None, overlapping=False):
        """
        Set the parameters of the integration class.

        Inputs:

            n        :    Number of coherent integrations
            timeInterval   :    Time of integration. If the parameter "n" is selected this one does not work
            overlapping    :

        """

        self.__initime = None
        self.__lastdatatime = 0

        self.__buffer_spc = 0
        self.__buffer_cspc = 0
        self.__buffer_dc = 0

        self.__profIndex = 0
        self.__dataReady = False
        self.__byTime = False

        if n is None and timeInterval is None:
            raise ValueError("n or timeInterval should be specified ...")

        if n is not None:
            self.n = int(n)
        else:
            
            self.__integrationtime = int(timeInterval)
            self.n = None
            self.__byTime = True

    def putData(self, data_spc, data_cspc, data_dc):
        """
        Add a profile to the __buffer_spc and increase in one the __profileIndex

        """

        self.__buffer_spc += data_spc

        if data_cspc is None:
            self.__buffer_cspc = None
        else:
            self.__buffer_cspc += data_cspc

        if data_dc is None:
            self.__buffer_dc = None
        else:
            self.__buffer_dc += data_dc

        self.__profIndex += 1

        return

    def pushData(self):
        """
        Return the sum of the last profiles and the profiles used in the sum.

        Affected:

        self.__profileIndex

        """

        data_spc = self.__buffer_spc
        data_cspc = self.__buffer_cspc
        data_dc = self.__buffer_dc
        n = self.__profIndex

        self.__buffer_spc = 0
        self.__buffer_cspc = 0
        self.__buffer_dc = 0
        self.__profIndex = 0

        return data_spc, data_cspc, data_dc, n

    def byProfiles(self, *args):

        self.__dataReady = False
        avgdata_spc = None
        avgdata_cspc = None
        avgdata_dc = None

        self.putData(*args)

        if self.__profIndex == self.n:

            avgdata_spc, avgdata_cspc, avgdata_dc, n = self.pushData()
            self.n = n
            self.__dataReady = True

        return avgdata_spc, avgdata_cspc, avgdata_dc

    def byTime(self, datatime, *args):

        self.__dataReady = False
        avgdata_spc = None
        avgdata_cspc = None
        avgdata_dc = None

        self.putData(*args)

        if (datatime - self.__initime) >= self.__integrationtime:
            avgdata_spc, avgdata_cspc, avgdata_dc, n = self.pushData()
            self.n = n
            self.__dataReady = True

        return avgdata_spc, avgdata_cspc, avgdata_dc

    def integrate(self, datatime, *args):

        if self.__profIndex == 0:
            self.__initime = datatime

        if self.__byTime:
            avgdata_spc, avgdata_cspc, avgdata_dc = self.byTime(
                datatime, *args)
        else:
            avgdata_spc, avgdata_cspc, avgdata_dc = self.byProfiles(*args)

        if not self.__dataReady:
            return None, None, None, None

        return self.__initime, avgdata_spc, avgdata_cspc, avgdata_dc

    def run(self, dataOut, n=None, timeInterval=None, overlapping=False):
        if n == 1:
            return dataOut
        
        dataOut.flagNoData = True

        if not self.isConfig:
            self.setup(n, timeInterval, overlapping)
            self.isConfig = True

        avgdatatime, avgdata_spc, avgdata_cspc, avgdata_dc = self.integrate(dataOut.utctime,
                                                                            dataOut.data_spc,
                                                                            dataOut.data_cspc,
                                                                            dataOut.data_dc)

        if self.__dataReady:

            dataOut.data_spc = avgdata_spc
            dataOut.data_cspc = avgdata_cspc
            dataOut.data_dc = avgdata_dc            
            dataOut.nIncohInt *= self.n
            dataOut.utctime = avgdatatime
            dataOut.flagNoData = False

        return dataOut


import numpy

from .jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.model.data.jrodata import SpectraHeis
from schainpy.utils import log



class SpectraHeisProc(ProcessingUnit):

    def __init__(self):#, **kwargs):

        ProcessingUnit.__init__(self)#, **kwargs)

#        self.buffer = None
#        self.firstdatatime = None
#        self.profIndex = 0
        self.dataOut = SpectraHeis()

    def __updateObjFromVoltage(self):

        self.dataOut.timeZone = self.dataIn.timeZone
        self.dataOut.dstFlag = self.dataIn.dstFlag
        self.dataOut.errorCount = self.dataIn.errorCount
        self.dataOut.useLocalTime = self.dataIn.useLocalTime

        self.dataOut.radarControllerHeaderObj = self.dataIn.radarControllerHeaderObj.copy()#
        self.dataOut.systemHeaderObj = self.dataIn.systemHeaderObj.copy()#
        self.dataOut.channelList = self.dataIn.channelList
        self.dataOut.heightList = self.dataIn.heightList
#        self.dataOut.dtype = self.dataIn.dtype
        self.dataOut.dtype = numpy.dtype([('real','<f4'),('imag','<f4')])
#        self.dataOut.nHeights = self.dataIn.nHeights
#        self.dataOut.nChannels = self.dataIn.nChannels
        self.dataOut.nBaud = self.dataIn.nBaud
        self.dataOut.nCode = self.dataIn.nCode
        self.dataOut.code = self.dataIn.code
#        self.dataOut.nProfiles = 1
        self.dataOut.ippFactor = 1
        self.dataOut.noise_estimation = None
#        self.dataOut.nProfiles = self.dataOut.nFFTPoints
        self.dataOut.nFFTPoints = self.dataIn.nHeights
#        self.dataOut.channelIndexList = self.dataIn.channelIndexList
#        self.dataOut.flagNoData = self.dataIn.flagNoData
        self.dataOut.flagDiscontinuousBlock = self.dataIn.flagDiscontinuousBlock
        self.dataOut.utctime = self.dataIn.utctime
#        self.dataOut.utctime = self.firstdatatime
        self.dataOut.flagDecodeData = self.dataIn.flagDecodeData #asumo q la data esta decodificada
        self.dataOut.flagDeflipData = self.dataIn.flagDeflipData #asumo q la data esta sin flip
#        self.dataOut.flagShiftFFT = self.dataIn.flagShiftFFT
        self.dataOut.nCohInt = self.dataIn.nCohInt
        self.dataOut.nIncohInt = 1
#         self.dataOut.ippSeconds= self.dataIn.ippSeconds
        self.dataOut.windowOfFilter = self.dataIn.windowOfFilter

#         self.dataOut.timeInterval = self.dataIn.timeInterval*self.dataOut.nIncohInt
#        self.dataOut.set=self.dataIn.set
#        self.dataOut.deltaHeight=self.dataIn.deltaHeight


    def __updateObjFromFits(self):

        self.dataOut.utctime = self.dataIn.utctime
#         self.dataOut.channelIndexList = self.dataIn.channelIndexList

        self.dataOut.channelList = self.dataIn.channelList
        self.dataOut.heightList = self.dataIn.heightList
        self.dataOut.data_spc = self.dataIn.data
        self.dataOut.ippSeconds = self.dataIn.ippSeconds
        self.dataOut.nCohInt = self.dataIn.nCohInt
        self.dataOut.nIncohInt = self.dataIn.nIncohInt
#         self.dataOut.timeInterval = self.dataIn.timeInterval
        self.dataOut.timeZone = self.dataIn.timeZone
        self.dataOut.useLocalTime = True
#         self.dataOut.
#         self.dataOut.

    def __getFft(self):

        fft_volt = numpy.fft.fft(self.dataIn.data, axis=1)
        fft_volt = numpy.fft.fftshift(fft_volt,axes=(1,))
        spc = numpy.abs(fft_volt * numpy.conjugate(fft_volt))/(self.dataOut.nFFTPoints)
        self.dataOut.data_spc = spc

    def run(self):

        self.dataOut.flagNoData = True

        if self.dataIn.type == "Fits":
            self.__updateObjFromFits()
            self.dataOut.flagNoData = False
            return 

        if self.dataIn.type == "SpectraHeis":
            self.dataOut.copy(self.dataIn)
            return

        if self.dataIn.type == "Voltage":
            self.__updateObjFromVoltage()
            self.__getFft()
            self.dataOut.flagNoData = False

            return

        raise ValueError("The type object %s is not valid"%(self.dataIn.type))


    def selectChannels(self, channelList):

        channelIndexList = []

        for channel in channelList:
            index = self.dataOut.channelList.index(channel)
            channelIndexList.append(index)

        self.selectChannelsByIndex(channelIndexList)

    def selectChannelsByIndex(self, channelIndexList):
        """
        Selecciona un bloque de datos en base a canales segun el channelIndexList

        Input:
            channelIndexList    :    lista sencilla de canales a seleccionar por ej. [2,3,7]

        Affected:
            self.dataOut.data
            self.dataOut.channelIndexList
            self.dataOut.nChannels
            self.dataOut.m_ProcessingHeader.totalSpectra
            self.dataOut.systemHeaderObj.numChannels
            self.dataOut.m_ProcessingHeader.blockSize

        Return:
            None
        """

        for channelIndex in channelIndexList:
            if channelIndex not in self.dataOut.channelIndexList:
                print(channelIndexList)
                raise ValueError("The value %d in channelIndexList is not valid" %channelIndex)

#         nChannels = len(channelIndexList)

        data_spc = self.dataOut.data_spc[channelIndexList,:]

        self.dataOut.data_spc = data_spc
        self.dataOut.channelList = [self.dataOut.channelList[i] for i in channelIndexList]

        return 1


class IncohInt4SpectraHeis(Operation):

    isConfig = False

    __profIndex = 0
    __withOverapping  = False

    __byTime = False
    __initime = None
    __lastdatatime = None
    __integrationtime = None

    __buffer = None

    __dataReady = False

    n = None

    def __init__(self):#, **kwargs):

        Operation.__init__(self)#, **kwargs)
#         self.isConfig = False

    def setup(self, n=None, timeInterval=None, overlapping=False):
        """
        Set the parameters of the integration class.

        Inputs:

            n        :    Number of coherent integrations
            timeInterval   :    Time of integration. If the parameter "n" is selected this one does not work
            overlapping    :

        """

        self.__initime = None
        self.__lastdatatime = 0
        self.__buffer = None
        self.__dataReady = False


        if n == None and timeInterval == None:
            raise ValueError("n or timeInterval should be specified ...")

        if n != None:
            self.n = n
            self.__byTime = False
        else:
            self.__integrationtime = timeInterval #* 60. #if (type(timeInterval)!=integer) -> change this line
            self.n = 9999
            self.__byTime = True

        if overlapping:
            self.__withOverapping = True
            self.__buffer = None
        else:
            self.__withOverapping = False
            self.__buffer = 0

        self.__profIndex = 0

    def putData(self, data):

        """
        Add a profile to the __buffer and increase in one the __profileIndex

        """

        if not self.__withOverapping:
            self.__buffer += data.copy()
            self.__profIndex += 1
            return

        #Overlapping data
        nChannels, nHeis = data.shape
        data = numpy.reshape(data, (1, nChannels, nHeis))

        #If the buffer is empty then it takes the data value
        if self.__buffer is None:
            self.__buffer = data
            self.__profIndex += 1
            return

        #If the buffer length is lower than n then stakcing the data value
        if self.__profIndex < self.n:
            self.__buffer = numpy.vstack((self.__buffer, data))
            self.__profIndex += 1
            return

        #If the buffer length is equal to n then replacing the last buffer value with the data value
        self.__buffer = numpy.roll(self.__buffer, -1, axis=0)
        self.__buffer[self.n-1] = data
        self.__profIndex = self.n
        return


    def pushData(self):
        """
        Return the sum of the last profiles and the profiles used in the sum.

        Affected:

        self.__profileIndex

        """

        if not self.__withOverapping:
            data = self.__buffer
            n = self.__profIndex

            self.__buffer = 0
            self.__profIndex = 0

            return data, n

        #Integration with Overlapping
        data = numpy.sum(self.__buffer, axis=0)
        n = self.__profIndex

        return data, n

    def byProfiles(self, data):

        self.__dataReady = False
        avgdata = None
#         n = None

        self.putData(data)

        if self.__profIndex == self.n:

            avgdata, n = self.pushData()
            self.__dataReady = True

        return avgdata

    def byTime(self, data, datatime):

        self.__dataReady = False
        avgdata = None
        n = None

        self.putData(data)

        if (datatime - self.__initime) >= self.__integrationtime:
            avgdata, n = self.pushData()
            self.n = n
            self.__dataReady = True

        return avgdata

    def integrate(self, data, datatime=None):

        if self.__initime == None:
            self.__initime = datatime

        if self.__byTime:
            avgdata = self.byTime(data, datatime)
        else:
            avgdata = self.byProfiles(data)


        self.__lastdatatime = datatime

        if avgdata is None:
            return None, None

        avgdatatime = self.__initime

        deltatime = datatime -self.__lastdatatime

        if not self.__withOverapping:
            self.__initime = datatime
        else:
            self.__initime += deltatime

        return avgdata, avgdatatime

    def run(self, dataOut, n=None, timeInterval=None, overlapping=False, **kwargs):

        if not self.isConfig:
            self.setup(n=n, timeInterval=timeInterval, overlapping=overlapping)
            self.isConfig = True

        avgdata, avgdatatime = self.integrate(dataOut.data_spc, dataOut.utctime)

#        dataOut.timeInterval *= n
        dataOut.flagNoData = True

        if self.__dataReady:
            dataOut.data_spc = avgdata
            dataOut.nIncohInt *= self.n
#            dataOut.nCohInt *= self.n
            dataOut.utctime = avgdatatime
#             dataOut.timeInterval = dataOut.ippSeconds * dataOut.nIncohInt
#            dataOut.timeInterval = self.__timeInterval*self.n
            dataOut.flagNoData = False
        
        return dataOut


import click
import subprocess
import os
import sys
import glob
import schainpy
from schainpy.controller import Project
from schainpy.model import Operation, ProcessingUnit
from schainpy.utils import log
from importlib import import_module
from pydoc import locate
from fuzzywuzzy import process
from schainpy.cli import templates
import inspect
try:
    from queue import Queue
except:
    from Queue import Queue


def getProcs():
    modules = dir(schainpy.model)
    procs = check_module(modules, 'processing')    
    try:
        procs.remove('ProcessingUnit')
    except Exception as e:
        pass
    return procs

def getOperations():
    module = dir(schainpy.model)
    noProcs = [x for x in module if not x.endswith('Proc')]
    operations = check_module(noProcs, 'operation')
    try:
        operations.remove('Operation')
        operations.remove('Figure')
        operations.remove('Plot')
    except Exception as e:
        pass
    return operations

def getArgs(op):
    module = locate('schainpy.model.{}'.format(op))
    try:
        obj = module(1, 2, 3, Queue())
    except:
        obj = module()

    if hasattr(obj, '__attrs__'):
        args = obj.__attrs__
    else:
        if hasattr(obj, 'myrun'):
            args = inspect.getfullargspec(obj.myrun).args
        else:
            args = inspect.getfullargspec(obj.run).args
    
    try:
        args.remove('self')
    except Exception as e:
        pass
    try:
        args.remove('dataOut')
    except Exception as e:
        pass
    return args

def getDoc(obj):    
    module = locate('schainpy.model.{}'.format(obj))
    try:
        obj = module(1, 2, 3, Queue())
    except:
        obj = module()
    return obj.__doc__

def getAll():
    modules = getOperations()
    modules.extend(getProcs())    
    return modules


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(schainpy.__version__)
    ctx.exit()


PREFIX = 'experiment'

@click.command()
@click.option('--version', '-v', is_flag=True, callback=print_version, help='SChain version', type=str)
@click.argument('command', default='run', required=True)
@click.argument('nextcommand', default=None, required=False, type=str)
def main(command, nextcommand, version):
    """COMMAND LINE INTERFACE FOR SIGNAL CHAIN - JICAMARCA RADIO OBSERVATORY V3.0\n
        Available commands:\n
        xml: runs a schain XML generated file\n
        run: runs any python script'\n
        generate: generates a template schain script\n
        list: return a list of available procs and operations\n
        search: return avilable operations, procs or arguments of the given
                operation/proc\n"""
    if command == 'xml':
        runFromXML(nextcommand)
    elif command == 'generate':
        generate()
    elif command == 'test':
        test()
    elif command == 'run':
        runschain(nextcommand)
    elif command == 'search':
        search(nextcommand)
    elif command == 'list':
        cmdlist(nextcommand)
    else:
        log.error('Command {} is not defined'.format(command))


def check_module(possible, instance):
    def check(x):
        try:
            instancia = locate('schainpy.model.{}'.format(x))
            ret = instancia.proc_type == instance
            return ret
        except Exception as e:
            return False
    clean = clean_modules(possible)
    return [x for x in clean if check(x)]


def clean_modules(module):
    noEndsUnder = [x for x in module if not x.endswith('__')]
    noStartUnder = [x for x in noEndsUnder if not x.startswith('__')]
    noFullUpper = [x for x in noStartUnder if not x.isupper()]
    return noFullUpper

def cmdlist(nextcommand):
    if nextcommand is None:
        log.error('Missing argument, available arguments: procs, operations', '')
    elif nextcommand == 'procs':
        procs = getProcs()
        log.success(
            'Current ProcessingUnits are:\n  {}'.format('\n  '.join(procs)), '')
    elif nextcommand == 'operations':
        operations = getOperations()
        log.success('Current Operations are:\n  {}'.format(
            '\n  '.join(operations)), '')
    else:
        log.error('Wrong argument', '')

def search(nextcommand):
    if nextcommand is None:
        log.error('There is no Operation/ProcessingUnit to search', '')    
    else:
        try:
            args = getArgs(nextcommand)
            doc = getDoc(nextcommand)
            log.success('{}\n{}\n\nparameters:\n  {}'.format(
                nextcommand, doc, ', '.join(args)), ''
                )
        except Exception as e:
            log.error('Module `{}` does not exists'.format(nextcommand), '')
            allModules = getAll()
            similar = [t[0] for t in process.extract(nextcommand, allModules, limit=12) if t[1]>80]
            log.success('Possible modules are: {}'.format(', '.join(similar)), '')

def runschain(nextcommand):
    if nextcommand is None:
        currentfiles = glob.glob('./{}_*.py'.format(PREFIX))
        numberfiles = len(currentfiles)
        if numberfiles > 1:
            log.error('There is more than one file to run')
        elif numberfiles == 1:
            subprocess.call(['python ' + currentfiles[0]], shell=True)
        else:
            log.error('There is no file to run')
    else:
        try:
            subprocess.call(['python ' + nextcommand], shell=True)
        except Exception as e:
            log.error("I cannot run the file. Does it exists?")


def basicInputs():
    inputs = {}
    inputs['name'] = click.prompt(
        'Name of the project', default="project", type=str)
    inputs['desc'] = click.prompt(
        'Enter a description', default="A schain project", type=str)
    inputs['multiprocess'] = click.prompt(
        '''Select data type: 

    - Voltage (*.r):                [1]
    - Spectra (*.pdata):            [2]
    - Voltage and Spectra (*.r):    [3]
    
    -->''', type=int)
    inputs['path'] = click.prompt('Data path', default=os.getcwd(
    ), type=click.Path(exists=True, resolve_path=True))
    inputs['startDate'] = click.prompt(
        'Start date', default='1970/01/01', type=str)
    inputs['endDate'] = click.prompt(
        'End date', default='2018/12/31', type=str)
    inputs['startHour'] = click.prompt(
        'Start hour', default='00:00:00', type=str)
    inputs['endHour'] = click.prompt('End hour', default='23:59:59', type=str)
    inputs['figpath'] = inputs['path'] + '/figs'
    return inputs


def generate():
    inputs = basicInputs()

    if inputs['multiprocess'] == 1:
        current = templates.voltage.format(**inputs)
    elif inputs['multiprocess'] == 2:
        current = templates.spectra.format(**inputs)
    elif inputs['multiprocess'] == 3:
        current = templates.voltagespectra.format(**inputs)
    scriptname = '{}_{}.py'.format(PREFIX, inputs['name'])
    script = open(scriptname, 'w')
    try:
        script.write(current)
        log.success('Script {} generated'.format(scriptname))
    except Exception as e:
        log.error('I cannot create the file. Do you have writing permissions?')


def test():
    log.warning('testing')


def runFromXML(filename):
    controller = Project()
    if not controller.readXml(filename):
        return
    controller.start()
    return
